"""
WHAD Enhanced ShockBurst Stack - Link-layer management

This module provides the ESB stack Link-layer management class
`LinkLayer`. This class uses its own associated state class
`LinkLayerState`.
"""
import logging
from queue import Queue, Empty
from time import sleep, time
from typing import Optional, Generator

from scapy.packet import Packet

from whad.esb.stack.llm.exceptions import LinkLayerTimeoutException
from whad.esb.stack.llm.constants import ESBRole
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response
from whad.common.stack import Layer, alias, source, state, LayerState, instance

logger = logging.getLogger(__name__)

class LinkLayerState(LayerState):
    """ESB Link-layer state class.

    This class stores the link-layer state.
    """

    def __init__(self):
        """Initialization of link-layer state.
        """
        super().__init__()

        self.pid = 0
        self.role = ESBRole.PTX

        self.synchronized = False
        self.promiscuous = False

        self.last_timestamp = time()

        self.ack_miss = 0


@alias('ll')
@state(LinkLayerState)
class LinkLayer(Layer):
    """ESB Link-layer management class.

    This class manages a single ESB connection and its state,
    as well as promiscuous mode.
    """

    def configure(self, options: Optional[dict] = None):
        """Configure this ESB link-layer instance.

        This method is called by the underlying layer management
        system to configure this layer.

        :param options: Layer options
        :type options: dict, optional
        """
        self.__ack_queue = Queue()
        self.__data_queue = Queue()

    @property
    def channel(self) -> int:
        """Return the current channel number

        :return: Current channel number
        :rtype: int
        """
        return self.get_layer('phy').channel

    @channel.setter
    def channel(self, channel: int):
        """Set the current channel number

        :param channel: New channel number to use
        :type channel: int
        """
        self.get_layer('phy').channel = channel

    @property
    def address(self) -> str:
        """Return the current ESB address.

        :return: Current ESB address
        :rtype: str
        """
        return self.get_layer('phy').address

    @address.setter
    def address(self, address: str):
        """Set ESB address

        :param address: New ESB address to use
        :type address: str
        """
        self.get_layer('phy').address = address

    @property
    def app(self):
        """Return the associated application layer instance

        :return: Associated application layer instance
        :rtype: Layer
        """
        return self.get_layer('app')

    @property
    def promiscuous(self) -> bool:
        """Determine if promiscuous mode is enabled or not

        :return: Promiscuous mode state
        :rtype: bool
        """
        return self.state.promiscuous

    @promiscuous.setter
    def promiscuous(self, promiscuous: bool):
        """Enable or disable promiscuous mode.

        :param promiscuous: Set to `True` to enable promiscuous mode,
                            `False` to disable.
        :type promiscuous: bool
        """
        self.state.promiscuous = promiscuous

    @property
    def synchronized(self) -> bool:
        """Determine if connection is synchronized or not.

        :return: Synchronization state
        :rtype: bool
        """
        return self.state.synchronized

    @property
    def role(self) -> int:
        """Get the current role.

        :return: Current role as defined in `ESBRole` constants.
        :rtype: int
        """
        return self.state.role

    @role.setter
    def role(self, role: int):
        """Set current role.

        :param role: New role
        :type role: int
        """
        self.state.role = role

    def _increment_pid(self):
        """Increment PID.
        """
        self.state.pid = (self.state.pid + 1) % 4

    def synchronize(self, timeout: float = 10.0):
        """Synchronize with current connection.

        :param timeout: Synchronization timeout
        :type timeout: float
        """
        self.state.role = ESBRole.PTX
        self.channel = None
        start_time = time()
        self.__ack_queue.queue.clear()
        self.__data_queue.queue.clear()
        self.state.promiscuous = True
        queue = self.__ack_queue
        while (time() - start_time) < timeout:
            try:
                queue = self.__ack_queue if queue == self.__data_queue else self.__data_queue
                msg = queue.get(block=False,timeout=0.05)

                if hasattr(msg, "metadata") and hasattr(msg.metadata, "channel"):
                    self.channel = msg.metadata.channel

                    if not self.state.synchronized:
                        self.state.synchronized = True
                        self.on_synchronized()
                    break
            except Empty:
                pass
        self.state.promiscuous = False
        return self.state.synchronized

    @source('app', 'data')
    def send_data(self, data: bytes, waiting_ack: Optional[bool] = False):
        """Send data to current connection.

        :param data: Data to send
        :type data: bytes
        :param waiting_ack: Set to `True` to wait for the remote device's
                            acknowledgement, `False` to skip.
        :type waiting_ack: bool, optional
        """
        self.state.role = ESBRole.PTX
        packet = ESB_Hdr(
                pid=self.state.pid,
                address=self.get_layer('phy').address,
                no_ack=0
        ) / ESB_Payload_Hdr() / data

        self.send('phy', packet, channel=self.get_layer('phy').channel)
        self._increment_pid()
        if waiting_ack:
            try:
                ack = self.wait_for_ack()
                self.state.ack_miss = 0
                self.state.synchronized = True
                return ack
            except LinkLayerTimeoutException:
                self.state.ack_miss += 1
                if self.state.ack_miss > 10:
                    self.state.ack_miss = 0
                    if self.state.synchronized:
                        self.state.synchronized = False
                        self.on_desynchronized()
        # No ack received
        return None

    def wait_for_ack(self, timeout: Optional[float] = 0.1):
        """Wait for an acknowledgement from remote device.

        :param timeout: Waiting timeout
        :type timeout: float, optional
        :raises: LinkLayerTimeoutException
        """
        self.state.role = ESBRole.PTX
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self.__ack_queue.get(block=False,timeout=0.001)
                return msg
            except Empty:
                pass
        raise LinkLayerTimeoutException

    def wait_for_data(self, timeout: Optional[float] = 0.1):
        """Wait for incoming data packet.

        :param timeout: Waiting timeout
        :type timeout: float, optional
        :raises: LinkLayerTimeoutException
        """
        self.state.role = ESBRole.PRX
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                msg = self.__data_queue.get(block=False,timeout=0.001)
                return msg
            except Empty:
                pass
        raise LinkLayerTimeoutException

    def data_stream(self) -> Generator[Packet, None, None]:
        """Generator that yields received data packets as they
        come.
        """
        self.state.role = ESBRole.PRX
        while True:
            yield self.__data_queue.get(block=True)

    @source('app', 'ack')
    def prepare_acknowledgment(self, data: bytes):
        """Send an acknowledgement for a given packet.

        :param data: Packet to acknowledge
        :type data: bytes
        """
        self.state.role = ESBRole.PRX
        packet = ESB_Hdr(
                pid=self.state.pid,
                address=self.get_layer('phy').address,
                no_ack=True
        ) / ESB_Payload_Hdr() / data
        self.send('phy', packet)

    def on_synchronized(self):
        """Synchronization callback that is called each time
        the stack has synchronized with the remote device.
        """
        if self.get_layer('app') is not None:
            self.send('app', time(), tag='synchronized')

    def on_desynchronized(self):
        """Desynchronization callback that is called each time
        the stack has desynchronized from the remote device.
        """
        if self.get_layer('app') is not None:
            self.send('app', time(), tag='desynchronized')


    def on_prx_pdu(self, pdu: Packet):
        """PRX packet reception callback.

        :param pdu: Packet received
        :type pdu: Packet
        """
        if (self.state.role == ESBRole.PTX or self.state.promiscuous):
            self.__ack_queue.put(pdu)

            if self.get_layer('app') is not None and len(bytes(pdu)) > 0:
                self.send('app', pdu[ESB_Payload_Hdr:], tag='ack')

    def on_ptx_pdu(self, pdu: Packet):
        """PTX packet reception callback.

        :param pdu: Packet transmitted
        :type pdu: Packet
        """
        if (self.state.role == ESBRole.PRX or self.state.promiscuous):

            self.__data_queue.put(pdu)

            if self.get_layer('app') is not None:
                self.send('app', pdu[ESB_Payload_Hdr:], tag='data')

    @source('phy')
    def on_pdu(self, pdu):
        """Packet reception callback.

        This callback dispatches the received packets to
        correct callbacks depending on the current mode.
        """
        if ESB_Ack_Response in pdu or len(bytes(pdu)) == 0:
            self.on_prx_pdu(pdu)
        else:
            self.on_ptx_pdu(pdu)



"""WHAD Protocol Discovery info messages abstraction layer.
"""
from whad.hub.message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper, \
    PbFieldArray
from whad.hub.discovery import Discovery

@pb_bind(Discovery, 'info_query', 1)
class InfoQuery(PbMessageWrapper):
    """Device info query message class
    """
    proto_ver = PbFieldInt('discovery.info_query.proto_ver')

@pb_bind(Discovery, 'info_resp', 1)
class InfoQueryResp(PbMessageWrapper):
    """Device info query response message class
    """
    type = PbFieldInt('discovery.info_resp.type')
    device_id = PbFieldBytes('discovery.info_resp.devid')
    proto_min_ver = PbFieldInt('discovery.info_resp.proto_min_ver')
    max_speed = PbFieldInt('discovery.info_resp.max_speed')
    fw_author = PbFieldBytes('discovery.info_resp.fw_author')
    fw_url = PbFieldBytes('discovery.info_resp.fw_url')
    fw_version_major = PbFieldInt('discovery.info_resp.fw_version_major')
    fw_version_minor = PbFieldInt('discovery.info_resp.fw_version_minor')
    fw_version_rev = PbFieldInt('discovery.info_resp.fw_version_rev')
    capabilities = PbFieldArray('discovery.info_resp.capabilities')



"""WHAD Protocol ESB domain message abstraction layer.
"""
from typing import List, Union
from dataclasses import dataclass, field, fields

from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr, ESB_Ack_Response, \
    ESB_Pseudo_Packet

from whad.hub.registry import Registry
from whad.hub.message import HubMessage, pb_bind
from whad.hub import ProtocolHub
from whad.hub.metadata import Metadata

class Commands:
    """ESB Commands
    """
    SetNodeAddress = 0x00
    Sniff = 0x01
    Jam = 0x02
    Send = 0x03
    SendRaw = 0x04
    PrimaryReceiverMode = 0x05
    PrimaryTransmitterMode = 0x06
    Start = 0x07
    Stop = 0x08

class EsbNodeAddressError(Exception):
    def __init__(self, message=''):
        super().__init__()
        self.__message = message

    def __repr__(self):
        return f'EsbNodeAddressError({self.__message})'

class EsbNodeAddress(object):
    """ESB node address

    This class is used to represent an ESB node address:

    >>> addr = EsbNodeAddress(0x1122334455, 5)
    >>> print(addr.value)
    >>> addr2 = EsbNodeAddress(bytes(0x11, 0x22, 0x33, 0x44, 0x55))
    >>> print(addr2.value)
    """

    def __init__(self, address: Union[str, bytes, int], address_size: int = 0):
        """Initialize our node address

        :param address: ESB node address
        :type address: bytes, int
        :param address_size: Node address size (1-5)
        :type address_size: int, optional
        """
        self.__address = None
        if isinstance(address, str):
            self.__address = bytes([int(i, 16) for i in address.split(":")])
        elif isinstance(address, bytes):
            if len(address) >= 1 and len(address) <= 5:
                self.__address = address
            else:
                raise EsbNodeAddressError('address size must be between 1 and 5 bytes long')
        elif isinstance(address, int) and address_size >= 1 and address_size <= 5:
            try:
                self.__address = address.to_bytes(address_size, 'big')
            except OverflowError as overflow_err:
                raise EsbNodeAddressError('address size must be big enough to hold address') from overflow_err
        else:
            raise EsbNodeAddressError('address must be provided as an array of bytes or an integer')

    @property
    def value(self) -> bytes:
        """Retrieve the address value (bytes)
        """
        return self.__address


    def __eq__(self, other) -> bool:
        """Compare two ESB node addresses
        """
        return self.value == other.value


@dataclass(repr=False)
class ESBMetadata(Metadata):
    is_crc_valid : bool = None
    address : str = None
    timestamp : int = None

    def convert_to_header(self):
        return None, self.timestamp

    @classmethod
    def convert_from_header(cls, pkt):
        metadata = ESBMetadata()
        pkt = ESB_Hdr(bytes(pkt))
        metadata.address = EsbNodeAddress(pkt.address)
        metadata.is_crc_valid = pkt.valid_crc
        metadata.timestamp = int(100000 * pkt.time)
        metadata.channel = 0
        return metadata

def generate_esb_metadata(message):
    metadata = ESBMetadata()

    if message.rssi is not None:
        metadata.rssi = message.rssi
    metadata.channel = message.channel
    if message.timestamp is not None:
        metadata.timestamp = message.timestamp
    if message.crc_validity is not None:
        metadata.is_crc_valid = message.crc_validity
    if message.address is not None:
        metadata.address = ":".join(["{:02x}".format(i) for i in message.address])
    return metadata

@pb_bind(ProtocolHub, name="esb", version=1)
class EsbDomain(Registry):
    """WHAD ESB domain messages parser/factory.
    """

    NAME = 'esb'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a ESB domain instance
        """
        self.proto_version = version
        from whad.scapy.layers.unifying import unbind
        unbind()

    def is_packet_compat(self, packet) -> bool:
        """Determine if a packet is an ESB packet.
        """
        return isinstance(packet.metadata, ESBMetadata)

    def convert_packet(self, packet) -> HubMessage:
        """Convert an ESB packet to SendPdu or SendBlePdu message.
        """
        if isinstance(packet.metadata, ESBMetadata):
            if packet.metadata.raw:
                return EsbDomain.bound('send_raw', self.proto_version).from_packet(
                    packet
                )
            else:
                return EsbDomain.bound('send', self.proto_version).from_packet(
                    packet
                )
        else:
            # Error
            return None

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if ESB_Hdr not in packet:
            if hasattr(packet, "metadata") and hasattr(packet.metadata, "address"):
                address = packet.metadata.address
            else:
                address = "11:22:33:44:55"

            packet = ESB_Hdr(address = address)/packet

        packet.preamble = 0xAA # force a rebuild
        formatted_packet = ESB_Pseudo_Packet(bytes(packet))

        timestamp = None
        if hasattr(packet, "metadata"):
            timestamp = packet.metadata.timestamp

        return formatted_packet, timestamp

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD ESB Domain message as seen by protobuf
        """
        message_type = message.esb.WhichOneof('msg')
        message_clazz = EsbDomain.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)


    def create_set_node_address(self, node_address: EsbNodeAddress) -> HubMessage:
        """Create a SetNodeAddress message

        :param node_address: Node address to set (size must be 1-5 bytes)
        :type node_address: EsbNodeAddress
        :return: instance of `SetNodeAddress`
        """
        return EsbDomain.bound('set_node_addr', self.proto_version)(
            address=node_address.value
        )

    def create_start(self) -> HubMessage:
        """Create a Start message

        :return: instance of `Start` message
        """
        return EsbDomain.bound('start', self.proto_version)()


    def create_stop(self) -> HubMessage:
        """Create a Stop message

        :return: instance of `Stop` message
        """
        return EsbDomain.bound('stop', self.proto_version)()


    def create_jam_mode(self, channel: int) -> HubMessage:
        """Create a JamMode message

        :param channel: ESB channel to jam
        :type channel: int
        :return: instance of `JamMode`
        """
        return EsbDomain.bound('jam', self.proto_version)(
            channel=channel
        )

    def create_sniff_mode(self, address: EsbNodeAddress, channel: int = 0xFF, show_acks: bool = False) -> HubMessage:
        """Create a SniffMode message

        :param address: Node address to filter
        :type address: EsbNodeAddress
        :param channel: Channel to sniff
        :type channel: int, optional
        :param show_acks: show acknowledgements
        :type show_acks: bool, optional
        :return: instance of `SniffMode`
        """
        return EsbDomain.bound('sniff', self.proto_version)(
            address=address.value,
            channel=channel,
            show_acks=show_acks
        )

    def create_jammed(self, timestamp: int):
        """Create a Jammed notification message

        :param timestamp: Timestamp at which the jamming has succeeded
        :type timestamp: int
        :return: instance of `Jammed`
        """
        return EsbDomain.bound('jammed', self.proto_version)(
            timestamp=timestamp
        )

    def create_prx_mode(self, channel: int) -> HubMessage:
        """Create PrxMode message

        :param channel: Channel to listen on
        :type channel: int
        :return: instance of `PrxMode`
        """
        return EsbDomain.bound('prx', self.proto_version)(
            channel=channel
        )

    def create_ptx_mode(self, channel: int) -> HubMessage:
        """Create PtxMode message

        :param channel: Channel to listen on
        :type channel: int
        :return: instance of `PtxMode`
        """
        return EsbDomain.bound('ptx', self.proto_version)(
            channel=channel
        )

    def create_send_pdu(self, channel: int, pdu: bytes, retr_count: int = 0):
        """Create a SendPdu message

        :param channel: Channel to use for transmission
        :type channel: int
        :param pdu: Data to send
        :type pdu: bytes
        :param retr_count: Retransmission count
        :type retr_count: int
        :return: instance of `SendPdu`
        """
        return EsbDomain.bound('send', self.proto_version)(
            channel=channel,
            pdu=pdu,
            retr_count=retr_count
        )

    def create_send_raw_pdu(self, channel: int, pdu: bytes, retr_count: int = 0):
        """Create a SendRawPdu message

        :param channel: Channel to use for transmission
        :type channel: int
        :param pdu: Raw data to send
        :type pdu: bytes
        :param retr_count: Retransmission count
        :type retr_count: int
        :return: instance of `SendPdu`
        """
        return EsbDomain.bound('send_raw', self.proto_version)(
            channel=channel,
            pdu=pdu,
            retr_count=retr_count
        )

    def create_pdu_received(self, channel: int, pdu: bytes, rssi: int = None, timestamp: int = None,
                          crc_validity: bool = None, address: EsbNodeAddress = None) -> HubMessage:
        """Create a PduReceived notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Data received (PDU)
        :type pdu: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Reception timestamp
        :type timestamp: int, optional
        :param crc_validity: Indicate CRC validity
        :type crc_validity: bool, optional
        :param address: Sender address
        :type address: EsbNodeAddress, optional
        :return: instance of `PduReceived`
        """
        # Create our base message
        msg = EsbDomain.bound('pdu', self.proto_version)(
            channel=channel,
            pdu=pdu
        )

        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if address is not None:
            msg.address = address.value
        if crc_validity is not None:
            msg.crc_validity = crc_validity

        # Return message
        return msg


    def create_raw_pdu_received(self, channel: int, pdu: bytes, rssi: int = None, timestamp: int = None,
                          crc_validity: bool = None, address: EsbNodeAddress = None) -> HubMessage:
        """Create a RawPduReceived notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Raw data received (PDU)
        :type pdu: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Reception timestamp
        :type timestamp: int, optional
        :param crc_validity: Indicate CRC validity
        :type crc_validity: bool, optional
        :param address: Sender address
        :type address: EsbNodeAddress, optional
        :return: instance of `PduReceived`
        """
        # Create our base message
        msg = EsbDomain.bound('raw_pdu', self.proto_version)(
            channel=channel,
            pdu=pdu
        )

        # Add optional fields if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if address is not None:
            msg.address = address.value
        if crc_validity is not None:
            msg.crc_validity = crc_validity

        # Return message
        return msg

from .address import SetNodeAddress
from .mode import EsbStart, EsbStop, JamMode, SniffMode, Jammed, PrxMode, PtxMode
from .pdu import PduReceived, RawPduReceived, SendPdu, SendRawPdu

__all__ = [
    'EsbDomain',
    'SetNodeAddress',
    'EsbStart',
    'EsbStop',
    'JamMode',
    'SniffMode',
    'Jammed',
    'PrxMode',
    'PtxMode',
    'SendPdu',
    'SendRawPdu',
    'PduReceived',
    'RawPduReceived'
]



"""
International HID keymaps
"""

from .be import HID_MAP as BE_MAP
from .ca import HID_MAP as CA_MAP
from .ch import HID_MAP as CH_MAP
from .de import HID_MAP as DE_MAP
from .dk import HID_MAP as DK_MAP
from .es import HID_MAP as ES_MAP
from .fi import HID_MAP as FI_MAP
from .fr import HID_MAP as FR_MAP
from .gb import HID_MAP as GB_MAP
from .hr import HID_MAP as HR_MAP
from .it import HID_MAP as IT_MAP
from .no import HID_MAP as NO_MAP
from .pt import HID_MAP as PT_MAP
from .ru import HID_MAP as RU_MAP
from .sl import HID_MAP as SL_MAP
from .sv import HID_MAP as SV_MAP
from .tr import HID_MAP as TR_MAP
from .us import HID_MAP as US_MAP

HID_MAP = {
    'be': BE_MAP,
    'ca': CA_MAP,
    'ch': CH_MAP,
    'de': DE_MAP,
    'dk': DK_MAP,
    'es': ES_MAP,
    'fi': FI_MAP,
    'fr': FR_MAP,
    'gb': GB_MAP,
    'hr': HR_MAP,
    'it': IT_MAP,
    'no': NO_MAP,
    'pt': PT_MAP,
    'ru': RU_MAP,
    'sl': SL_MAP,
    'sv': SV_MAP,
    'tr': TR_MAP,
    'us': US_MAP,
}

__all__ = [
    "HID_MAP",
    "BE_MAP",
    "CA_MAP",
    "CH_MAP",
    "DE_MAP",
    "DK_MAP",
    "ES_MAP",
    "FI_MAP",
    "FR_MAP",
    "GB_MAP",
    "HR_MAP",
    "IT_MAP",
    "NO_MAP",
    "PT_MAP",
    "RU_MAP",
    "SL_MAP",
    "SV_MAP",
    "TR_MAP",
    "US_MAP",
]



"""WHAD Protocol PHY timestamp message abstraction layer.
"""
from ..message import PbFieldInt, PbMessageWrapper

class Timestamp(PbMessageWrapper):
    """PHY timestamp message
    """
    sec = PbFieldInt('sec')
    usec = PbFieldInt('usec')


from enum import IntEnum
"""
Constants implemented as Enum for Zigbee NWK layer.
"""

class ZigbeeDeviceType(IntEnum):
    """
    Enum representing the different device types in a Zigbee network.
    """
    COORDINATOR = 0
    ROUTER = 1
    END_DEVICE = 2

class ZigbeeRelationship(IntEnum):
    """
    Enum representing the different relationship between ZigBee devices.
    """
    IS_PARENT = 0
    IS_CHILD = 1
    IS_SIBLING = 2
    NONE = 3
    PREVIOUS_CHILD = 4
    UNAUTHENTICATED_CHILD = 5


class NWKJoinMode(IntEnum):
    """
    Enum representing the different NWK join modes supported by 802.15.4.
    """
    NEW_JOIN = 0
    REJOIN = 1
    CHANGE_CHANNEL = 2

class NWKAddressMode(IntEnum):
    """
    Enum representing the different NWK address modes supported by 802.15.4.
    """
    MULTICAST = 1
    UNICAST = BROADCAST = 2

BROADCAST_ADDRESSES = {
    0xFFFF : "All devices in PAN",
    0xFFFD : "macRxOnWhenIdle=True",
    0xFFFC : "All routers and coordinators",
    0xFFFB : "Low power routers only"
}



from whad.zigbee.stack.apl.zdo.object import ZDOObject
from whad.zigbee.stack.apl.zdo.descriptors import NodeDescriptor, SimpleDescriptor
from whad.zigbee.profile.nodes import CoordinatorNode, EndDeviceNode, RouterNode
from whad.dot15d4.stack.mac.constants import MACDeviceType, MACPowerSource
from whad.zigbee.stack.apl.constants import LogicalDeviceType
from whad.zigbee.stack.apl.zdo.discovery.exceptions import ZDODeviceAndServiceDiscoveryTimeoutException
from whad.scapy.layers.zdp import ZDPIEEEAddrReq, ZDPSimpleDescRsp, ZDPActiveEPReq, ZDPActiveEPRsp, \
    ZDPNodeDescRsp, ZDPIEEEAddrRsp, ZDPSimpleDescReq, ZDPNodeDescReq
from whad.zigbee.stack.nwk.constants import ZigbeeRelationship, ZigbeeDeviceType
from queue import Queue, Empty
from time import time, sleep

import logging

logger = logging.getLogger(__name__)

class ZDODeviceAndServiceDiscovery(ZDOObject):
    """
    ZDO Device Object handling the discovery of Devices, Apps & Services.
    """

    def __init__(self, zdo):
        # Init transaction counter
        self.transaction = 0
        # Build input queue
        self.input_queue = Queue()

        super().__init__(zdo)


    def can_transmit(self):
        """
        Check if we are associated and authorized to transmit on the network
        """

        return (
            self.zdo.manager.get_layer("nwk").database.get("nwkNetworkAddress") != 0xFFFF and
            self.zdo.network_manager.authorized
        )



    def get_simple_descriptor(self, node_address, endpoint):
        """
        Discover the simple descriptor of a specific endpoint for a specific node.
        """
        logger.info("[zdo_device_and_service_discovery_manager] Discovering simple descriptor of endpoint #%s of node %s.",str(endpoint), hex(node_address))

        # If we are not allowed to transmit, trigger an error
        if not self.can_transmit():
            logger.info("[zdo_device_and_service_discovery_manager] Simple descriptor discovery failure, no associated and authorized network.")
            return None

        # Send a Simple Descriptor Request
        self.zdo.clusters["simple_desc_req"].send_data(
            node_address,
            endpoint,
            self.transaction
        )
        self.transaction += 1

        # Get Simple Descriptor Response or trigger an error if timeout
        try:
            response = self.wait_for_response(filter_function=
                lambda pkt:ZDPSimpleDescRsp in pkt# and pkt.trans_seqnum == self.transaction
            )
            (asdu, source_address, source_address_mode, security_status, link_quality) = response

            # If successful, return the associated simple descriptor
            if asdu.status == 0:
                return SimpleDescriptor(
                        endpoint=asdu.endpoint,
                        profile_identifier=asdu.profile_identifier,
                        device_identifier=asdu.device_identifier,
                        device_version=asdu.device_version,
                        input_clusters=asdu.input_clusters,
                        output_clusters=asdu.output_clusters
                )
            return None

        except ZDODeviceAndServiceDiscoveryTimeoutException:
            return None


    def get_active_endpoints(self, node_address):
        """
        Discover the active endpoints exposed by a specific node.
        """
        logger.info("[zdo_device_and_service_discovery_manager] Discovering active endpoints of node %s.", hex(node_address))

        # If we are not allowed to transmit, trigger an error
        if not self.can_transmit():
            logger.info("[zdo_device_and_service_discovery_manager] Active endpoints discovery failure, no associated and authorized network.")
            return None

        # Send an Active Endpoint Request
        self.zdo.clusters["active_ep_req"].send_data(
            node_address,
            self.transaction
        )
        self.transaction += 1

        # Get Active Endpoint Response or trigger an error if timeout
        try:
            response = self.wait_for_response(
                filter_function=lambda pkt:ZDPActiveEPRsp in pkt# and pkt.trans_seqnum == self.transaction
            )
            (asdu, source_address, source_address_mode, security_status, link_quality) = response


            # If successful, return the list of active endpoints
            if asdu.status == 0:
                return asdu.active_endpoints
            return []

        except ZDODeviceAndServiceDiscoveryTimeoutException:
            return None


    def get_node_descriptor(self, node_address):
        """
        Discover node descriptor of a specific node.
        """
        logger.info("[zdo_device_and_service_discovery_manager] Discovering descriptor of node %s.", hex(node_address))

        # If we are not allowed to transmit, trigger an error
        if not self.can_transmit():
            logger.info("[zdo_device_and_service_discovery_manager] Node Descriptor discovery failure, no associated and authorized network.")
            return None

        # Send a Node Descriptor Request
        self.zdo.clusters["node_desc_req"].send_data(
            node_address,
            self.transaction
        )
        self.transaction += 1

        # Get Node Descriptor Response or trigger an error if timeout
        try:
            response = self.wait_for_response(
                filter_function=lambda pkt:ZDPNodeDescRsp in pkt# and pkt.trans_seqnum == self.transaction
            )
            (asdu, source_address, source_address_mode, security_status, link_quality) = response

            # If response indicates a successful statuts, return the Node Descriptor
            if asdu.status == 0:

                nd = NodeDescriptor(
                    logical_type=LogicalDeviceType(asdu.logical_type),
                    complex_descriptor_available=bool(asdu.complex_descriptor_available),
                    user_descriptor_available=bool(asdu.user_descriptor_available),
                    aps_flags=asdu.aps_flags,
                    support_868_mhz=bool(asdu.support_868_mhz),
                    support_902_mhz=bool(asdu.support_902_mhz),
                    support_2400_mhz=bool(asdu.support_2400_mhz),
                    alternate_pan_coordinator=bool(asdu.alternate_pan_coordinator),
                    device_type=MACDeviceType(asdu.device_type),
                    power_source=MACPowerSource(asdu.power_source),
                    receiver_on_when_idle=bool(asdu.receiver_on_when_idle),
                    security_capability=bool(asdu.security_capability),
                    allocate_address=bool(asdu.allocate_address),
                    manufacturer_code=asdu.manufacturer_code,
                    max_buffer_size=asdu.max_buffer_size,
                    max_incoming_transfer_size=asdu.max_incoming_transfer_size,
                    server_primary_trust_center = bool(asdu.server_primary_trust_center),
                    server_backup_trust_center = bool(asdu.server_backup_trust_center),
                    server_primary_binding_table_cache = bool(asdu.server_primary_binding_table_cache),
                    server_backup_binding_table_cache = bool(asdu.server_backup_binding_table_cache),
                    server_primary_discovery_cache = bool(asdu.server_primary_discovery_cache),
                    server_backup_discovery_cache = bool(asdu.server_backup_discovery_cache),
                    network_manager = bool(asdu.network_manager),
                    stack_compliance_revision = asdu.stack_compliance_revision,
                    max_outgoing_transfer_size = asdu.max_outgoing_transfer_size,
                    extended_active_endpoint_list_available = bool(asdu.extended_active_endpoint_list_available),
                    extended_simple_descriptors_list_available = bool(asdu.extended_simple_descriptors_list_available)
                )
                return nd
            return None

        except ZDODeviceAndServiceDiscoveryTimeoutException:
            return None

    def device_annce(self, transaction=0):
        """
        Annonce the current node.
        """
        self.zdo.clusters["device_annce"].send_data(transaction)


    def discover_nodes(self):
        """
        Discover nodes present on the current network.
        """
        logger.info("[zdo_device_and_service_discovery_manager] Discovering nodes.")
        # If we are not allowed to transmit, trigger an error
        if not self.can_transmit():
            logger.info("[zdo_device_and_service_discovery_manager] Nodes discovery failure, no associated network.")
            return []

        # Build a list of nodes addresses in the network
        network = self.zdo.network_manager.network
        addresses = [device.address for device in network.nodes]

        # Send an IEEE Address Request in broadcast
        self.zdo.clusters["ieee_addr_req"].send_data(
            0xFFFF,
            request_type=1,
            transaction=self.transaction
        )
        self.transaction += 1

        # Wait for responses and populate addresses list if we find more nodes
        try:
            while True:
                response = self.wait_for_response(
                    filter_function=lambda pkt:ZDPIEEEAddrRsp in pkt# and pkt.trans_seqnum == self.transaction
                )
                (asdu, source_address, source_address_mode, security_status, link_quality) = response

                if source_address not in addresses:
                    addresses.append(source_address)

        except ZDODeviceAndServiceDiscoveryTimeoutException:
            pass

        # Iterate over addresses in our list
        scanned_nodes = []
        for address in addresses:
            new_device = None
            # Send an IEEE Address Request

            self.zdo.clusters["ieee_addr_req"].send_data(
                address,
                request_type=1,
                transaction=self.transaction
            )
            self.transaction += 1
            try:
                response = self.wait_for_response(
                    filter_function=lambda pkt:ZDPIEEEAddrRsp in pkt# and pkt.trans_seqnum == self.transaction
                )
                (asdu, source_address, source_address_mode, security_status, link_quality) = response
                # Get the node descriptor and generate an appropriate node wrapper
                descriptor = self.get_node_descriptor(address)

                if descriptor is not None:
                    if descriptor.logical_type == LogicalDeviceType.COORDINATOR:

                        new_device = CoordinatorNode(
                            address,
                            extended_address=asdu.ieee_addr,
                            descriptor=descriptor,
                            network=network
                        )
                        network.coordinator = new_device
                    elif descriptor.logical_type == LogicalDeviceType.ROUTER:
                        new_device = RouterNode(
                            address,
                            extended_address=asdu.ieee_addr,
                            descriptor=descriptor,
                            network=network
                        )
                        if new_device not in network.routers:
                            network.routers.append(new_device)
                    else:
                        new_device = EndDeviceNode(
                            address,
                            extended_address=asdu.ieee_addr,
                            descriptor=descriptor,
                            network=network
                        )
                        if new_device not in network.end_devices:
                            network.end_devices.append(new_device)

                    scanned_nodes.append(new_device)

            except ZDODeviceAndServiceDiscoveryTimeoutException:
                pass
        return scanned_nodes

    def on_cluster_data(
                        self,
                        asdu,
                        source_address,
                        source_address_mode,
                        security_status,
                        link_quality
    ):
        """
        Callback called when a cluster processes a PDU.
        """
        # Send to queue
        self.input_queue.put(
            (
                asdu,
                source_address,
                source_address_mode,
                security_status,
                link_quality
            )
        )

        if ZDPIEEEAddrReq in asdu:
            self.on_ieee_addr_req(asdu, source_address)
        elif ZDPActiveEPReq in asdu:
            self.on_active_ep_req(asdu, source_address)
        elif ZDPSimpleDescReq in asdu:
            self.on_simple_desc_req(asdu, source_address)
        elif ZDPNodeDescReq in asdu:
            self.on_node_desc_req(asdu, source_address)

    def on_node_desc_req(self, asdu, source_address):
        """
        Callback called when a Node Descriptor Request is received.
        """
        nwk_layer = self.zdo.manager.get_layer("nwk")
        own_network_address = nwk_layer.database.get("nwkNetworkAddress")

        apl_layer = self.zdo.manager.get_layer("apl")


        associated_devices_addresses = []
        for address, device in nwk_layer.database.get("nwkNeighborTable").table.items():
            if (
                device.relationship == ZigbeeRelationship.IS_CHILD and
                device.device_type == ZigbeeDeviceType.END_DEVICE
            ):
                associated_devices_addresses.append(device.address)


        if own_network_address == asdu.nwk_addr:
            node_descriptor = self.zdo.configuration.configNodeDescriptor
            # Send a Node Descriptor Request
            self.zdo.clusters["node_desc_rsp"].send_data(
                address,
                status=0,
                node_descriptor=node_descriptor,
                transaction=self.transaction
            )
            self.transaction += 1
        elif asdu.nwk_addr in associated_devices_addresses:
            self.zdo.clusters["node_desc_rsp"].send_data(
                address,
                status=3,
                node_descriptor=node_descriptor,
                transaction=self.transaction
            )
            self.transaction += 1
        else:
            self.zdo.clusters["node_desc_rsp"].send_data(
                address,
                status=1,
                node_descriptor=node_descriptor,
                transaction=self.transaction
            )
            self.transaction += 1

    def on_simple_desc_req(self, asdu, remote_address):
        """
        Callback called when a Simple Descriptor Request is received.
        """
        nwk_layer = self.zdo.manager.get_layer("nwk")
        own_network_address = nwk_layer.database.get("nwkNetworkAddress")

        apl_layer = self.zdo.manager.get_layer("apl")


        associated_devices_addresses = []
        for address, device in nwk_layer.database.get("nwkNeighborTable").table.items():
            if (
                device.relationship == ZigbeeRelationship.IS_CHILD and
                device.device_type == ZigbeeDeviceType.END_DEVICE
            ):
                associated_devices_addresses.append(device.address)


        if own_network_address == asdu.nwk_addr:
            # It matches our own address, check if endpoint in active endpoints
            endpoints = apl_layer.get_endpoints()

            if asdu.endpoint in endpoints and asdu.endpoint != 0:
                app = apl_layer.get_application_by_endpoint(asdu.endpoint)
                simple_descriptor = app.simple_descriptor

                self.zdo.clusters["simple_desc_rsp"].send_data(
                    status = 0, # success
                    local_address = own_network_address,
                    remote_address = remote_address,
                    descriptor = simple_descriptor,
                    transaction = self.transaction
                )
                self.transaction += 1
            else:
                self.zdo.clusters["simple_desc_rsp"].send_data(
                    status = 3, # no descriptor
                    local_address = own_network_address,
                    remote_address = remote_address,
                    transaction = self.transaction
                )
                self.transaction += 1
        elif asdu.nwk_addr in associated_devices_addresses:
            # It matches one of our children
            self.zdo.clusters["simple_desc_rsp"].send_data(
                status = 3, # no descriptor (we lie for now)
                local_address = own_network_address,
                remote_address = remote_address,
                transaction = self.transaction
            )
            self.transaction += 1
        else:
            self.zdo.clusters["simple_desc_rsp"].send_data(
                status = 1, # device not found
                local_address = own_network_address,
                remote_address = remote_address,
                transaction = self.transaction
            )
            self.transaction += 1


    def on_active_ep_req(self, asdu, remote_address):
        """
        Callback called when an active endpoint request is received.
        """
        nwk_layer = self.zdo.manager.get_layer("nwk")
        own_network_address = nwk_layer.database.get("nwkNetworkAddress")

        apl_layer = self.zdo.manager.get_layer("apl")

        if asdu.nwk_addr == own_network_address:
            endpoints = apl_layer.get_endpoints()
            endpoints.remove(0)
            self.zdo.clusters["active_ep_rsp"].send_data(
                status = 0,
                local_address = own_network_address,
                remote_address = remote_address,
                endpoints = endpoints,
                transaction = self.transaction
            )
            self.transaction += 1

        # Not implemented: discovery of sub devices in the discovery type
        else:
            self.zdo.clusters["active_ep_rsp"].send_data(
                status = 1, # dev not found
                local_address = asdu.nwk_addr,
                remote_address = remote_address,
                endpoints = [],
                transaction = self.transaction
            )
            self.transaction += 1


    def on_ieee_addr_req(self, asdu, remote_address):
        """
        Callback called when a IEEE address request is received.
        """
        nwk_layer = self.zdo.manager.get_layer("nwk")
        own_network_address = nwk_layer.database.get("nwkNetworkAddress")
        own_ieee_address = nwk_layer.database.get("nwkIeeeAddress")


        associated_devices_addresses = []
        for address, device in nwk_layer.database.get("nwkNeighborTable").table.items():
            if (
                device.relationship == ZigbeeRelationship.IS_CHILD and
                device.device_type == ZigbeeDeviceType.END_DEVICE
            ):
                associated_devices_addresses.append(device.address)


        match = False
        if own_network_address == asdu.nwk_addr or asdu.nwk_addr in associated_devices_addresses:
            match = True


        if match:
            if asdu.request_type == 0:
                self.zdo.clusters["ieee_addr_rsp"].send_data(
                    own_network_address,
                    own_ieee_address,
                    remote_address = remote_address,
                    status=0,
                    num_assoc_dev=0,
                    start_index=0,
                    associated_devices=[],
                    transaction=self.transaction
                )
            elif asdu.request_type == 1:
                self.zdo.clusters["ieee_addr_rsp"].send_data(
                    own_network_address,
                    own_ieee_address,
                    remote_address = remote_address,
                    status=0,
                    num_assoc_dev=len(associated_devices_addresses),
                    start_index=asdu.start_index,
                    associated_devices=associated_devices_addresses[asdu.start_index:],
                    transaction=self.transaction
                )
            else:
                self.zdo.clusters["ieee_addr_rsp"].send_data(
                    own_network_address,
                    own_ieee_address,
                    remote_address = remote_address,
                    status=2,
                    num_assoc_dev=0,
                    start_index=0,
                    associated_devices=[],
                    transaction=self.transaction
                )

            self.transaction += 1

        else:
            # no match, send device not found response
            self.zdo.clusters["ieee_addr_rsp"].send_data(
                own_network_address,
                own_ieee_address,
                remote_address = remote_address,
                status=1,
                num_assoc_dev=0,
                start_index=0,
                associated_devices=[],
                transaction=self.transaction
            )
            self.transaction += 1

    def wait_for_response(self, filter_function=lambda pkt:True, timeout=3):
        """
        Synchronous blocking method for processing data from input ZDP clusters.
        """
        start_time = time()
        while (time() - start_time) < timeout:
            try:
                (
                    asdu,
                    source_address,
                    source_address_mode,
                    security_status,
                    link_quality
                ) = self.input_queue.get(block=False,timeout=0.1)
                if filter_function(asdu):
                    return (
                        asdu,
                        source_address,
                        source_address_mode,
                        security_status,
                        link_quality
                    )
            except Empty:
                pass

        raise ZDODeviceAndServiceDiscoveryTimeoutException



"""WHAD Protocol BLE address messages abstraction layer.
"""

from ..message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper
from . import BleDomain

@pb_bind(BleDomain, 'set_bd_addr', 1)
class SetBdAddress(PbMessageWrapper):
    """Device info query message class
    """

    bd_address = PbFieldBytes('ble.set_bd_addr.bd_address')
    addr_type = PbFieldInt('ble.set_bd_addr.addr_type')



"""Bluetooth Low Energy Device Information Service Profile
"""
from whad.ble.profile.attribute import UUID
from whad.ble.profile.service import StandardService
from whad.ble.profile.characteristic import Characteristic

class DeviceInformationService(StandardService):
    """Device Information Service version 1.2 as defined in
    `specification <https://www.bluetooth.com/specifications/specs/html/?src=DIS_v1.2/out/en/index-en.html>`_.
    """
    _uuid = UUID(0x180a)

    manufacturer= Characteristic(
        UUID(0x2A29),
        permissions=['read'],
    )
    model_number=Characteristic(
        UUID(0x2A24),
        permissions=['read']
    )
    serial_number=Characteristic(
        UUID(0x2A25),
        permissions=['read']
    )
    hw_revision=Characteristic(
        UUID(0x2A27),
        permissions=['read']
    )
    fw_revision=Characteristic(
        UUID(0x2A26),
        permissions=['read']
    )
    sw_revision=Characteristic(
        UUID(0x2A28),
        permissions=['read']
    )
    system_id=Characteristic(
        UUID(0x2A23),
        permissions=['read']
    )
    ieee_rcdl=Characteristic(
        UUID(0x2A2A),
        permissions=['read']
    )
    pnp_id=Characteristic(
        UUID(0x2A50),
        permissions=['read']
    )
    udi=Characteristic(
        UUID(0x2BFF),
        permissions = ['read'],
    )




"""BLE Packet translator
"""
import struct
import logging

from scapy.packet import Packet
from scapy.compat import raw
from scapy.layers.bluetooth4LE import BTLE, BTLE_ADV, BTLE_DATA, BTLE_ADV_IND, \
    BTLE_ADV_NONCONN_IND, BTLE_ADV_DIRECT_IND, BTLE_ADV_SCAN_IND, BTLE_SCAN_RSP, \
    BTLE_RF, BTLE_CTRL

from whad.hub.ble import generate_ble_metadata
from whad.hub import ProtocolHub
from whad.hub.message import HubMessage
from whad.hub.ble import AdvType, BleAdvPduReceived, BlePduReceived, BleRawPduReceived, \
    SendBlePdu, SendBleRawPdu

logger = logging.getLogger(__name__)

def packet_to_bytes(packet):
    """Convert packet to bytes
    """
    try:
        return raw(packet)
    except TypeError:
        return bytes(packet)

class BleMessageTranslator:
    """BLE Whad message translator.

    This translator is used to provide the format of a specific scapy packet
    as well as standard methods to convert WHAD BLE messages into scapy packets
    (if it makes sense) and scapy packets into WHAD BLE messages.
    """

    # correlation table
    SCAPY_CORR_ADV = {
        AdvType.ADV_IND: BTLE_ADV_IND,
        AdvType.ADV_NONCONN_IND: BTLE_ADV_NONCONN_IND,
        AdvType.ADV_DIRECT_IND: BTLE_ADV_DIRECT_IND,
        AdvType.ADV_SCAN_IND: BTLE_ADV_SCAN_IND,
        AdvType.ADV_SCAN_RSP: BTLE_SCAN_RSP
    }

    def __init__(self, protocol_hub: ProtocolHub):
        self.__access_address = 0x11223344
        self.__hub = protocol_hub


    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        formatted_packet = packet

        if BTLE not in packet:
            if BTLE_ADV in packet:
                formatted_packet = BTLE(access_addr=0x8e89bed6)/packet
            elif BTLE_DATA in packet:
                # We are forced to use a pseudo access address for connections in this case.
                formatted_packet = BTLE(access_addr=self.__access_address) / packet

        timestamp = None
        if hasattr(packet, "metadata"):
            header, timestamp = packet.metadata.convert_to_header()

            formatted_packet = header / formatted_packet
        else:
            header = BTLE_RF()
            formatted_packet = header / formatted_packet

        return formatted_packet, timestamp


    def from_message(self, message) -> Packet:
        """Convert a WHAD message into a packet, if it makes sense.
        """
        try:
            # Advertising PDU (RX)
            if isinstance(message, BleAdvPduReceived):
                if message.adv_type in BleMessageTranslator.SCAPY_CORR_ADV:
                    data = bytes(message.adv_data)

                    packet = BTLE_ADV()/BleMessageTranslator.SCAPY_CORR_ADV[message.adv_type](
                            bytes(message.bd_address) + data
                        )
                    packet.metadata = generate_ble_metadata(message)
                    return packet

            # Raw PDU (RX)
            elif isinstance(message, BleRawPduReceived):
                packet = BTLE(bytes(struct.pack("I", message.access_address)) + bytes(message.pdu) + bytes(struct.pack(">I", message.crc)[1:]))
                packet.metadata = generate_ble_metadata(message)
                return packet

            # Normal PDU (RX)
            elif isinstance(message, BlePduReceived):
                packet = BTLE_DATA(message.pdu)
                packet.metadata = generate_ble_metadata(message)
                return packet

            # Send PDU (TX)
            elif isinstance(message, SendBlePdu):
                packet = BTLE_DATA(message.pdu)
                packet.metadata = generate_ble_metadata(message)
                return packet

            # Send Raw PDU (TX)
            elif isinstance(message, SendBleRawPdu):
                packet = BTLE(bytes(struct.pack("I", message.access_address)) + bytes(message.pdu) + bytes(struct.pack(">I", message.crc)[1:]))
                packet.metadata = generate_ble_metadata(message)
                return packet

            # No matching packet found, return None.
            return None

        except AttributeError as err:
            logger.error(err)
            return None


    def from_packet(self, packet: Packet, encrypt: bool = False) -> HubMessage:
        """Convert a BLE PDU scapy packet into the corresponding hub message.
        """
        direction = packet.metadata.direction
        connection_handle = packet.metadata.connection_handle

        if BTLE in packet:
            # Extract PDU
            if BTLE_DATA in packet:
                pdu = raw(packet[BTLE_DATA:])
            elif BTLE_CTRL in packet:
                pdu = raw(packet[BTLE_CTRL:])
            elif BTLE_ADV in packet:
                pdu = raw(packet[BTLE_ADV:])
            else:
                return None

            # Create SendRawPdu message
            msg = self.__hub.ble.create_send_raw_pdu(
                direction,
                pdu,
                crc=BTLE(raw(packet)).crc, # force the CRC to be generated if not provided
                access_address=BTLE(raw(packet)).access_addr,
                conn_handle=connection_handle,
                encrypt=encrypt
            )

        else:

            # Extract PDU
            if BTLE_DATA in packet:
                pdu = packet_to_bytes(packet[BTLE_DATA:])
            elif BTLE_CTRL in packet:
                pdu = packet_to_bytes(packet[BTLE_CTRL:])
            elif BTLE_ADV in packet:
                pdu = packet_to_bytes(packet[BTLE_ADV:])
            else:
                return None

            # Create a SendPdu message
            msg = self.__hub.ble.create_send_pdu(
                direction,
                pdu,
                connection_handle,
                encrypt=encrypt
            )

        return msg



"""BLE Interactive shell
"""
from whad.cli.app import command
from whad.ble.cli.central.shell import BleCentralShell

@command('interactive')
def interactive_handler(app, _):
    """interactive BLE shell

    <ansicyan><b>interactive</b></ansicyan>

    Starts an interactive shell and let you interact with a BLE WHAD device:
    - scan devices
    - connect to a device
    - list services and characteristics
    - read/write characteristics
    """
    # Launch an interactive shell
    myshell = BleCentralShell(app.interface)
    myshell.run()



"""
WHAD Enhanced ShockBurst connectors.

This module provides a set of connectors that implement different roles. 
"""
from whad.esb.connector.base import ESB
from whad.esb.connector.scanner import Scanner
from whad.esb.connector.sniffer import Sniffer
from whad.esb.connector.prx import PRX
from whad.esb.connector.ptx import PTX

__all__ = [
    "ESB",
    "Scanner",
    "Sniffer",
    "PRX",
    "PTX",
]



"""
BLE Scanner connector
=====================

This module provides a scanner connector :class:`whad.ble.connector.scanner.Scanner`
for WHAD BLE devices, allowing to discover available BLE devices.

It implements two different scanning approaches:

* The first one is based on a *normal device scanning* based on WHAD BLE protocol:
  it puts the hardware adapter in device discovery mode and grab each advertisement
  received.

* The second one puts the hardware adapter in *sniffing mode* and sniffs advertisements
  sent on default channels. This mode is only available with WHAD devices that support
  sniffing.

This connector must be instantiated with a compatible WHAD device, as shown below:

.. code-block:: python

    device = WhadDevice.create('uart0')
    scanner = Scanner(device)

If the underlying device does not support scanning, this connector will raise
an :class:`UnsupportedCapability` exception.

"""
from time import time
from typing import Iterator, List, Optional

from scapy.packet import Packet
from scapy.layers.bluetooth4LE import BTLE_ADV

from whad.hub.ble import BleAdvPduReceived, BleRawPduReceived
from whad.exceptions import UnsupportedCapability, WhadDeviceNotReady

from .base import BLE
from ..scanning import AdvertisingDevicesDB, AdvertisingDevice

class Scanner(BLE):
    """
    BLE Observer interface for compatible WHAD device.
    """

    domain = 'ble'

    def __init__(self, device):
        """Instantiate scanner connector over `device`.

        :param  device: BLE WHAD device instance
        :type   device: :class:`whad.device.WhadDevice`
        """
        super().__init__(device)
        self.__db = AdvertisingDevicesDB()

        # Check device accept scanning mode
        if not self.can_scan():
            # Does our device accept sniffing mode and is able of sniffing advertising reports ?
            if not self.can_sniff_advertisements():
                raise UnsupportedCapability('Scan')
            # Stop current mode
            self.stop()

            # Set advertisement sniffing
            self.sniff_advertisements()
        else:
            # Stop current mode
            self.stop()

            # Enable active scanning
            self.enable_scan_mode(True)

    def start(self) -> bool:
        """Start the BLE scanner.

        Calling this method resets the discovered devices database and put
        the WHAD device into BLE scanning mode.
        """
        self.__db.reset()
        return super().start()

    def stop(self) -> bool:
        """Stop scanning.

        Stop scanning for devices, disable scan mode if used.
        """
        # Stop scanning
        result = super().stop()

        # Disable active scan mode (passive mode by default)
        if self.can_scan():
            self.enable_scan_mode(False)

        # Send result
        return result

    def __enter__(self) -> 'Scanner':
        """Special method used to use this connector as a context manager."""
        # Make sure we are started
        if not self.started:
            if not self.start():
                raise WhadDeviceNotReady()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Special method used to close the connector when a
        context manager is done with it."""
        if self.started:
            if not self.stop():
                raise WhadDeviceNotReady()

    def discover_devices(self, minimal_rssi = None, filter_address = None,
                         timeout: Optional[float] = None, updates: bool = False) -> Iterator[AdvertisingDevice]:
        """
        Parse incoming advertisements and yield discovered devices. If `updates`
        is set to `True`, all discovered devices will be reported with updated
        information.

        :param  minimal_rssi:       Minimal RSSI level
        :type   minimal_rssi:       float, optional
        :param  filter_address:     BD address of a device to discover
        :type   filter_address:     :class:`whad.ble.bdaddr.BDAddress`, optional
        :param  timeout:            Timeout in seconds
        :type   timeout:            float, optional
        :param  updates:            Enable/disable reporting updated information
                                    about already discovered devices
        :type   updates:            bool
        """

        # Start sniffing
        start_time = time()
        for advertisement in self.sniff(timeout=timeout):
            if minimal_rssi is None or advertisement.metadata.rssi > minimal_rssi:
                devices = self.__db.on_device_found(
                    advertisement.metadata.rssi,
                    advertisement,
                    filter_addr=filter_address,
                    updates=updates
                )

                for device in devices:
                    yield device

            if (timeout is not None) and (time() - start_time > timeout):
                break

    def sniff(self, messages: Optional[List] = None, timeout: Optional[float] = None) -> Iterator[Packet]:
        """
        Listen and yield incoming advertising PDUs.
        """
        # Make sure the hardware interface is started
        if not self.started:
            if not self.start():
                raise WhadDeviceNotReady()
            stop_on_exit = True
        else:
            stop_on_exit = False

        # Determine the type of message we are expecting to receive
        if self.support_raw_pdu():
            message_type = BleRawPduReceived
        else:
            message_type = BleAdvPduReceived

        # Loop until timeout reached or stopped
        start_anchor = time()
        while True:
            # Switch to sniffing mode
            for message in super().sniff(messages=(message_type), timeout=timeout):
                # Convert message from rebuilt PDU
                packet = message.to_packet()
                if packet is not None:
                    self.monitor_packet_rx(packet)
                    # Force TxAdd value to propagate the address type
                    if isinstance(message, BleAdvPduReceived):
                        if message.addr_type > 0:
                            packet.getlayer(BTLE_ADV).TxAdd = 1
                    yield packet

                # Make sure we did not run over a configured timeout, if so
                # exit the main while loop
                if timeout is not None:
                    if (time() - start_anchor) >= timeout:
                        break

        # Stop current mode if we enabled it
        if stop_on_exit:
            if not self.stop():
                raise WhadDeviceNotReady()

    def clear(self):
        """
        Clear device database.
        """
        self.__db.reset()




"""WHAD BLE GATT default services

This module provides different default GATT services as defined in
* https://www.bluetooth.com/specifications/specs/battery-service/
* https://www.bluetooth.com/specifications/specs/device-information-service/
"""
from whad.ble.profile.services.bas import BatteryService
from whad.ble.profile.services.dis import DeviceInformationService
from whad.ble.profile.services.gap import GapService
from whad.ble.profile.services.hr import HeartRateService

__all__ = [
    'GapService',
    'BatteryService',
    'HeartRateService',
    'DeviceInformationService'
]



from whad.zigbee.stack.nwk.constants import ZigbeeDeviceType
from whad.zigbee.profile.endpoint import Endpoint

class EndpointsDiscoveryException(Exception):
    """
    Default exception raised when an error occurred during Endpoints discovery.
    """
    def __init__(self):
        super().__init__()


class Node:
    """
    Class representing a Zigbee Node.
    """
    def __init__(
                    self,
                    device_type,
                    address,
                    extended_address=None,
                    descriptor=None,
                    network=None
    ):
        self.__device_type = device_type
        self.__address = address
        self.__extended_address = extended_address
        self.__node_descriptor = descriptor
        self.__active_endpoints = None
        self.__network = network

    def discover(self):
        """
        Discover the endpoints and the node descriptor exposed by this node.
        """
        for endpoint in self.endpoints:
            _ = endpoint.descriptor
        return (self.endpoints, self.descriptor)

    @property
    def descriptor(self):
        """
        Returns the node descriptor.
        """
        if (
            self.__node_descriptor is None and
            self.__network.is_associated() and
            self.__network.is_authorized()
        ):
            self.__node_descriptor = self.stack.get_layer('apl').get_application_by_name("zdo").device_and_service_discovery.get_node_descriptor(self.__address)
        return self.__node_descriptor

    @property
    def endpoints(self):
        """
        Returns the list of active endpoints.
        """
        if (
            self.__active_endpoints is None and
            self.__network.is_associated() and
            self.__network.is_authorized()
        ):
            self.__active_endpoints = []
            endpoints = self.stack.get_layer('apl').get_application_by_name("zdo").device_and_service_discovery.get_active_endpoints(self.__address)
            
            if endpoints is None:
                self.__active_endpoints = None
                raise EndpointsDiscoveryException()

            for endpoint in endpoints:
                self.__active_endpoints.append(Endpoint(endpoint, self))

        return self.__active_endpoints

    @property
    def address(self):
        """
        Returns the node short address.
        """
        return self.__address

    @property
    def extended_address(self):
        """
        Returns the node extended address.
        """
        return self.__extended_address

    @address.setter
    def address(self, address):
        """
        Sets the node short address.
        """
        self.__address = address


    @extended_address.setter
    def extended_address(self, extended_address):
        """
        Sets the node extended address.
        """
        self.__extended_address = extended_address

    @property
    def network(self):
        """
        Returns the network where the node is associated.
        """
        return self.__network

    @property
    def stack(self):
        """
        Returns the stack instance linked to the node.
        """
        return self.__network.stack

    def __eq__(self, other):
        """
        Checks if two nodes are equals (according to their short address).
        """
        return self.address == other.address

    def __repr__(self):
        return (
            self.__class__.__name__ + "(" +
                "address="+hex(self.address) + ", " +
                "extended_address="+(hex(self.extended_address) if self.extended_address is not None else "<unknown>") + ", " +
                "network="+hex(self.network.extended_pan_id) +
            ")"
        )
class CoordinatorNode(Node):
    """
    Represents a ZigBee Coordinator node.
    """
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    descriptor=None,
                    network=None
    ):
        super().__init__(
                            ZigbeeDeviceType.COORDINATOR,
                            address,
                            extended_address=extended_address,
                            descriptor=descriptor,
                            network=network
        )
class RouterNode(Node):
    """
    Represents a ZigBee Router node.
    """
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    descriptor=None,
                    network=None
    ):
        super().__init__(
                            ZigbeeDeviceType.ROUTER,
                            address,
                            extended_address=extended_address,
                            descriptor=descriptor,
                            network=network
        )

class EndDeviceNode(Node):
    """
    Represents a ZigBee End Device node.
    """
    def __init__(
                    self,
                    address,
                    extended_address=None,
                    descriptor=None,
                    network=None
    ):
        super().__init__(
                            ZigbeeDeviceType.END_DEVICE,
                            address,
                            extended_address=extended_address,
                            descriptor=descriptor,
                            network=network
        )



"""
PCAP capabilities module.

This module defines the PCAP virtual device capabilities for each supported
DLT.
"""
from whad.hub.discovery import Domain, Capability
from whad.hub.ble import Commands as BleCommands
from whad.hub.dot15d4 import Commands as Dot15d4Commands
from whad.hub.esb import Commands as EsbCommands
from whad.hub.unifying import Commands as UnifyingCommands
from whad.hub.phy import Commands as PhyCommands

# Bluetooth LE
DLT_BLUETOOTH_LE_LL_WITH_PHDR   = 256

# IEEE 802.15.4
DLT_IEEE802_15_4_LINUX          = 191
DLT_IEEE802_15_4_WITHFCS        = 195
DLT_IEEE802_15_4_NONASK_PHY     = 215
DLT_IEEE802_15_4_NOFCS          = 230
DLT_IEEE802_15_4_TAP            = 283

# WHAD ESB
DLT_RESERVED_02                 = 148
DLT_RESERVED_03                 = 149

# WHAD PHY
DLT_RESERVED_06                 = 152

CAPABILITIES = {
    # Bluetooth Low Energy
    DLT_BLUETOOTH_LE_LL_WITH_PHDR : (
        {
            Domain.BtLE : (
                (Capability.Sniff),
                [
                    BleCommands.SniffConnReq, BleCommands.SniffAdv,
                    BleCommands.Start, BleCommands.Stop
                ]
            )
        },
        {
            Domain.BtLE : (
                (Capability.Inject),
                [BleCommands.SendPDU, BleCommands.Start, BleCommands.Stop]
            )
        }
    ),

    # IEEE 802.15.4 - Various DLTs
    DLT_IEEE802_15_4_NONASK_PHY: (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),
    DLT_IEEE802_15_4_LINUX: (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),
    DLT_IEEE802_15_4_NOFCS: (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),
    DLT_IEEE802_15_4_WITHFCS: (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),
    DLT_IEEE802_15_4_TAP : (
        {
            Domain.Dot15d4 : (
                (Capability.Sniff),
                [Dot15d4Commands.Sniff, Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        },
        {
            Domain.Dot15d4 : (
                (Capability.Inject),
                [Dot15d4Commands.Send,Dot15d4Commands.Start, Dot15d4Commands.Stop]
            )
        }
    ),

    # Enhanced ShockBurst
    DLT_RESERVED_02 : (
        {
            Domain.Esb : (
                (Capability.Sniff),
                [EsbCommands.Sniff, EsbCommands.Start, EsbCommands.Stop]
            ),
            Domain.LogitechUnifying : (
                (Capability.Sniff),
                [UnifyingCommands.Sniff, UnifyingCommands.Start, UnifyingCommands.Stop]
            )
        },
        {
            Domain.Esb : (
                (Capability.Inject),
                [EsbCommands.Send,EsbCommands.Start, EsbCommands.Stop]
            ),
            Domain.LogitechUnifying : (
                (Capability.Inject),
                [UnifyingCommands.Send,UnifyingCommands.Start, UnifyingCommands.Stop]
            )
        }
    ),

    # WHAD PHY
    DLT_RESERVED_06 : (
        {
            Domain.Phy : (
                (Capability.Sniff | Capability.NoRawData),
                [
                    PhyCommands.Sniff, PhyCommands.Start, PhyCommands.Stop,
                    PhyCommands.SetFrequency, PhyCommands.GetSupportedFrequencies,
                    PhyCommands.SetPacketSize, PhyCommands.SetDataRate,
                    PhyCommands.SetEndianness, PhyCommands.SetSyncWord,
                    PhyCommands.SetASKModulation, PhyCommands.SetFSKModulation,
                    PhyCommands.SetGFSKModulation, PhyCommands.SetBPSKModulation,
                    PhyCommands.SetQPSKModulation, PhyCommands.SetLoRaModulation
                ]
            )
        },
        {
            Domain.Phy : (
                (Capability.Inject | Capability.NoRawData),
                [
                    PhyCommands.Send,PhyCommands.Start, PhyCommands.Stop,
                    PhyCommands.SetFrequency, PhyCommands.GetSupportedFrequencies,
                    PhyCommands.SetPacketSize, PhyCommands.SetDataRate,
                    PhyCommands.SetEndianness, PhyCommands.SetSyncWord,
                    PhyCommands.SetASKModulation, PhyCommands.SetFSKModulation,
                    PhyCommands.SetGFSKModulation, PhyCommands.SetBPSKModulation,
                    PhyCommands.SetQPSKModulation, PhyCommands.SetLoRaModulation
                ]
            )
        }
    )
}



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import sys
import logging
from time import sleep

# We need all scapy layers to craft filter expressions
# pylint: disable-next=wildcard-import
from scapy.layers.all import *
from scapy.packet import Packet

# We also need our custom layers
from whad.scapy.layers import *

# Required to configure scapy theme.
from scapy.themes import BrightTheme
from scapy.config import conf

from whad.cli.app import CommandLineDualOutputApp, run_app
from whad.device.unix import UnixSocketServer, UnixConnector
from whad.device import Bridge
from whad.hub import ProtocolHub
from whad.cli.ui import display_packet, error

logger = logging.getLogger(__name__)

class WhadFilterPipe(Bridge):
    """Whad filter output pipe

    When wfilter is chained with another whad tool, it spawns a device
    based on the specified profile using the specified WHAD adapter and forward
    it to the chained tool. The chained tool will then forward packets back and forth.
    """

    def __init__(self, input_connector, output_connector, on_rx_packet_cb, on_tx_packet_cb):
        super().__init__(input_connector, output_connector)
        self.on_rx_packet = on_rx_packet_cb
        self.on_tx_packet = on_tx_packet_cb


    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkt = self.on_rx_packet(pkt)
                if pkt is not None:
                    msg = message.from_packet(pkt)
                    super().on_outbound(msg)
        else:
            logger.debug("[wfilter][input-pipe] forward default outbound message %s", message)
            # Forward other messages
            super().on_outbound(message)


    def on_inbound(self, message):
        """Process inbound messages.

        Inbound packets are packets coming from our output connector,that need to be
        forwarded as packets to the previous tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkt = self.on_tx_packet(pkt)
                if pkt is not None:
                    msg = message.from_packet(pkt)
                    super().on_inbound(msg)
        else:
            logger.debug("[wfilter][input-pipe] forward default inbound message %s", message)
            # Forward other messages
            super().on_inbound(message)


class WhadFilterApp(CommandLineDualOutputApp):
    """wfilter CLI application class.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD filter tool',
            interface=True,
            commands=False,
            input=CommandLineDualOutputApp.INPUT_WHAD,
            output=CommandLineDualOutputApp.OUTPUT_WHAD
        )

        self.add_argument(
            'filter',
            help='filter to evaluate',
            nargs='?'
        )

        # Metada can be disabled here
        self.add_argument(
            '--no-metadata',
            dest='metadata',
            action="store_false",
            help='Hide packets metadata'
        )

        # Format can be overriden here
        self.add_argument(
            '--format',
            dest='format',
            action="store",
            default='repr',
            choices=['repr', 'show', 'raw', 'hexdump', 'tshark'],
            help='Indicate format to display packet'
        )

        self.add_argument(
            '--down',
            dest='down',
            action="store_true",
            default=None,
            help='process down stream'
        )


        self.add_argument(
            '--up',
            dest='up',
            action="store_true",
            default=None,
            help='process up stream'
        )

        self.add_argument(
            '-t',
            '--transform',
            dest='transform',
            type=str,
            default=None,
            help='apply a transformation'
        )

        self.add_argument(
            '-e',
            '--invert',
            dest='invert',
            action="store_true",
            default=False,
            help='invert filter'
        )

        self.add_argument(
            '-f',
            '--forward',
            dest='forward',
            action="store_true",
            default=False,
            help='forward packets not matched by the filter (dropped by default)'
        )

        self.add_argument(
            '-l',
            '--load',
            dest='loadables',
            default=None,
            action="append",
            help='load Scapy packet definitions from external Python file'
        )

        # Initialize our packet filter
        self.packet_filter = None

    def build_filter(self):
        """Build filter from provided args.
        """
        if self.args.filter is None:
            self.args.filter = "True"

        # Generate our packet filter lambda.
        if self.args.invert:
            filter_lambda = f"lambda p,pkt,packet: not {self.args.filter}"
        else:
            filter_lambda = f"lambda p,pkt,packet: {self.args.filter}"

        # Create our lambda function.
        try:
            # pylint: disable-next=eval-used
            return eval(filter_lambda)
        except SyntaxError:
            return None


    def on_rx_packet(self, pkt):
        """Packet reception callback.

        We apply our filter function to each packet received to determine if
        we must keep it or not. If no filter function is provided, we keep all
        of them.
        """
        if not self.args.down:
            if not self.is_stdout_piped():
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
            return pkt

        try:
            # We call our packet filter with the same arguments 3 times to
            # cover the `p`, `pkt` and `packet` variables in our lambda.
            if self.packet_filter(pkt, pkt, pkt):
                if self.args.transform is not None:

                    # Build our custom environment.
                    env = dict(globals())
                    env.update({
                        'p': pkt,
                        'pkt': pkt,
                        'packet': pkt
                    })

                    # Execute our transform.
                    # pylint: disable-next=exec-used
                    exec(self.args.transform, env)

                    # recompute CRC ?
                if not self.is_stdout_piped():
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                return pkt

            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt

            # No packet to forward.
            return None

        # We catch all types of exception here as the python code used by
        # the user may cause any issue.
        #
        # pylint: disable-next=broad-exception-caught
        except Exception:
            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt

            # No packet to forward.
            return None

    def on_tx_packet(self, pkt):
        """Packet transmission callback.
        """
        if not self.args.up:
            if not self.is_stdout_piped():
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
            return pkt

        try:
            # We call our packet filter with the same arguments 3 times to
            # cover the `p`, `pkt` and `packet` variables in our lambda.
            if self.packet_filter(pkt, pkt, pkt):
                if self.args.transform is not None:
                    # Build our custom environment.
                    env = dict(globals())
                    env.update({
                        'p': pkt,
                        'pkt': pkt,
                        'packet': pkt
                    })

                    # Execute our transform.
                    # pylint: disable-next=exec-used
                    exec(self.args.transform, env)

                if not self.is_stdout_piped():
                    display_packet(
                        pkt,
                        show_metadata = self.args.metadata,
                        format = self.args.format
                    )
                return pkt

            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt
            return None

        # We catch all types of exception here as the python code used by
        # the user may cause any issue.
        #
        # pylint: disable-next=broad-exception-caught
        except Exception:
            if self.args.forward:
                display_packet(
                    pkt,
                    show_metadata = self.args.metadata,
                    format = self.args.format
                )
                return pkt
            return None

    def run(self):
        """wfilter main function.
        """
        # Launch pre-run tasks
        self.pre_run()

        if self.args.down is None and self.args.up is None:
            self.args.up   = True
            self.args.down = True

        # Load any Scapy definition files if provided
        if self.args.loadables is not None:
            for loadable in self.args.loadables:
                l = __import__(loadable)
                for obj in dir(l):
                    o = getattr(l, obj)
                    try:
                        if issubclass(o, Packet) and o != Packet:
                            globals()[obj] = o
                    except TypeError:
                        pass

        try:
            # Build our packet filter, exit if invalid.
            self.packet_filter = self.build_filter()
            if self.packet_filter is None:
                error("Syntax error in provided filter expression.")
                return

            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                # Load parameters from input tool
                parameters = self.args.__dict__
                connector = UnixConnector(interface)

                # Set the current domain in the hub and inject into connector.
                ProtocolHub.set_domain(self.args.domain)
                connector.domain = self.args.domain

                if self.is_stdout_piped():
                    unix_server = UnixConnector(UnixSocketServer(parameters=parameters))


                    while not unix_server.device.opened:
                        if unix_server.device.timedout:
                            return
                        sleep(0.1)

                    # Create our packet bridge
                    logger.info("[wfilter] Starting our output pipe")
                    WhadFilterPipe(connector, unix_server, self.on_rx_packet,
                                   self.on_tx_packet).join()

                else:
                    # Unlock Unix connector
                    connector.unlock()

                    # Take format and metadata settings from input tool,
                    # if provided.
                    if "format" in parameters and parameters["format"] in ('repr', 'show',
                                                                           'raw', 'hexdump',
                                                                           'tshark'):
                        self.args.format = parameters["format"]

                    # Overwrite its packet rx method
                    connector.on_packet = self.on_rx_packet

                    # Keep running while interface is active
                    connector.join()
            else:
                sys.exit(1)
        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wfilter_main():
    """Launcher for wfilter CLI application.
    """
    app = WhadFilterApp()
    run_app(app)



"""Bluetooth Low Energy MitM tool

This utility will connect to a device and spawn a similar device that will act
as a proxy to the target device.
"""
import logging

from hexdump import hexdump
from prompt_toolkit import print_formatted_text, HTML

from whad.ble import Scanner, BDAddress
from whad.device import WhadDevice
from whad.ble.tools.proxy import GattProxy, LinkLayerProxy
from whad.cli.app import CommandLineDeviceSource, run_app
from whad.hub.ble import Direction

logger = logging.getLogger(__name__)

class VerboseLLProxy(LinkLayerProxy):
    """Verbose link-layer proxy class.

    This class intercepts any BLE link-layer operation and displays it in terminal.
    """

    def on_connect(self):
        """Handle connection
        """
        print_formatted_text(HTML(
            "<ansimagenta>Remote device connected</ansimagenta>"
        ))

    def on_disconnect(self):
        """Handle disconnection.
        """
        print_formatted_text(HTML(
            "<ansimagenta>Remote device disconnected</ansimagenta>"
        ))

    def on_ctl_pdu(self, pdu, direction):
        """Display captured Control PDU"""
        if direction == Direction.MASTER_TO_SLAVE:
            print_formatted_text(HTML("&lt;&lt;&lt; <ansicyan>Control PDU</ansicyan>"))
        else:
            print_formatted_text(HTML("&gt;&gt;&gt; <ansicyan>Control PDU</ansicyan>"))
        hexdump(bytes(pdu))
        return super().on_ctl_pdu(pdu, direction)


    def on_data_pdu(self, pdu, direction):
        """Display captured data PDU"""
        if direction == Direction.MASTER_TO_SLAVE:
            print_formatted_text(HTML("&lt;&lt;&lt; <ansimagenta>Data PDU</ansimagenta>"))
        else:
            print_formatted_text(HTML("&gt;&gt;&gt; <ansimagenta>Data PDU</ansimagenta>"))
        hexdump(bytes(pdu))
        return super().on_data_pdu(pdu, direction)

class VerboseProxy(GattProxy):
    """Main BLE GATT proxy
    """

    def on_characteristic_read(self, service, characteristic, value, offset=0, length=0):
        """Triggered when a characteristic read.
        """
        if offset > 0:
            print_formatted_text(HTML((
                f"&lt;&lt;&lt; <ansicyan>Characteristic {characteristic.uuid} read "
                f"(offset: {offset})</ansicyan>"
            )))
        else:
            print_formatted_text(HTML(
                f"&gt;&gt;&gt; <ansicyan>Characteristic {characteristic.uuid} read</ansicyan>"
            ))
        hexdump(value)

    def on_characteristic_write(self, service, characteristic, offset=0, value=b'',
                                without_response=False):
        """Triggered when a characteristic write.
        """
        if offset > 0:
            print_formatted_text(HTML((
                f"&lt;&lt;&lt; <ansicyan>Characteristic {characteristic.uuid} written "
                f"(offset: {offset})</ansicyan>"
            )))
        else:
            print_formatted_text(HTML(
                f"&gt;&gt;&gt; <ansicyan>Characteristic {characteristic.uuid} written</ansicyan>"
            ))
        hexdump(value)

    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        """Triggered when a GATT client subscribes to a characteristic.
        """
        if notification:
            print_formatted_text(HTML((
                    f"[!] <ansicyan>Subscribed to notification for charac. "
                    f"{characteristic.uuid}</ansicyan>"
            )))
        if indication:
            print_formatted_text(HTML((
                    f"[!] <ansicyan>Subscribed to indication for charac. "
                    f"{characteristic.uuid}</ansicyan>"
            )))

    def on_characteristic_unsubscribed(self, service, characteristic):
        """Triggered when a GATT client unsubscribes from a characteristic.
        """
        print_formatted_text(HTML(
                f"[!] <ansicyan>Unubscribed from charac. {characteristic.uuid}</ansicyan>"
        ))

    def on_notification(self, service, characteristic, value):
        """Triggered when a notification is received.
        """
        print_formatted_text(HTML((
            f"&lt;&lt;&lt; <ansicyan>[!] Notification for charac. "
            f"{characteristic.uuid}:</ansicyan>"
        )))
        hexdump(value)

    def on_indication(self, service, characteristic, value):
        """Triggered when a indication is received.
        """
        print_formatted_text(HTML((
            f"&lt;&lt;&lt; <ansicyan>[!] Indication for charac. "
            f"{characteristic.uuid}:</ansicyan>"
        )))
        hexdump(value)

    def on_connect(self, conn_handle):
        """Triggered when a remote device connects to our spoofed device.
        """
        super().on_connect(conn_handle)

        print_formatted_text(HTML(
            "<ansimagenta>Remote device connected</ansimagenta>"
        ))

    def on_disconnect(self, conn_handle):
        """Triggered when a remote device disconnects from our spoofed device.
        """
        print_formatted_text(HTML(
            "<ansimagenta>Remote device disconnected</ansimagenta>"
        ))

    def on_mtu_changed(self, mtu):
        print_formatted_text(HTML(
            f"<ansicyan>MTU changed to {mtu}"
        ))

class BleProxyApp(CommandLineDeviceSource):
    """Bluetooth Low-Energy Proxy application

    This application provides a GATT proxy feature that allows the user to connect
    to a target device, discovers its services and characteristics and spawns a
    duplicate device and proxify any BLE operation to the target device.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy proxy tool',
            interface=True,
            commands=False
        )

        self.add_argument("bdaddr", metavar="BDADDR", help="Target device BD address")

        # Add an optional random type argument
        self.add_argument(
            "-p",
            "--proxy-interface",
            dest="proxy_iface",
            type=str,
            required=True,
            help="Specify the WHAD interface to use for our spoofed device"
        )

        self.add_argument(
            "-t",
            "--timeout",
            dest="timeout",
            type=int,
            default=30,
            help="Scan timeout in seconds"
        )

        self.add_argument(
            "-w",
            "--wireshark",
            dest="wireshark",
            action="store_true",
            default=False,
            help="Enable real-time wireshark monitoring"
        )

        self.add_argument(
            "-s",
            "--spoof",
            dest="spoof",
            action="store_true",
            default=False,
            help="Enable BD address spoofing (if available)"
        )

        self.add_argument(
            "--link-layer",
            dest="linklayer",
            action="store_true",
            default=False,
            help="Enable link-layer mode"
        )

        self.add_argument(
            "-o",
            "--output",
            dest="output",
            default=None,
            help="Output PCAP file path"
        )

        self.proxy = None

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                if self.args.bdaddr is None:
                    self.error("Please provide a target BD address.")
                else:
                    self.spawn_proxy()
            else:
                self.error("You need to specify an interface with option --interface.")

        except KeyboardInterrupt:
            self.warning("wble-proxy stopped (CTL-C)")
            
            # Stop gracefully
            if self.proxy is not None:
                self.proxy.stop()

        # Launch post-run tasks
        self.post_run()

    def spawn_proxy(self):
        """Create a GATT proxy
        """
        # Exit if proxy interface not specified
        if self.args.proxy_iface is None:
            return

        proxy_iface = WhadDevice.create(self.args.proxy_iface)
        target = None

        # Start scanning, we are looking for our target device
        print(f"Scanning for target device (timeout: {self.args.timeout} seconds)...")
        scanner = Scanner(self.interface)
        scanner.start()
        for device in scanner.discover_devices(timeout=self.args.timeout):
            if device.address.lower() == self.args.bdaddr.lower():
                if device.adv_records is not None and device.scan_rsp_records is not None:
                    target = device
                    #adv_data = device.adv_records.to_bytes()
                    #scan_rsp = device.scan_rsp_records.to_bytes()
                    break

        # Device search timeout reached, show warning and stop proxy
        if target is None:
            self.warning("Target device not found, connection timeout exceeded.")
            scanner.stop()
            return

        # Stop scanning
        scanner.stop()

        # Display target device info
        addr_type = "random" if target.address_type == BDAddress.RANDOM else "public"
        print(f"Found target device {target.address} ({addr_type})")

        if target is not None:
            if not self.args.linklayer:
                self.proxy = VerboseProxy(
                    proxy_iface,
                    self.interface,
                    adv_data=target.adv_records.to_bytes(),
                    scan_data=target.scan_rsp_records.to_bytes(),
                    bd_address=self.args.bdaddr.lower(),
                    spoof=self.args.spoof,
                    random=(target.address_type == BDAddress.RANDOM)
                )
            else:
                self.proxy = VerboseLLProxy(
                    proxy=proxy_iface,
                    target=self.interface,
                    adv_data=target.adv_records.to_bytes(),
                    scan_data=target.scan_rsp_records.to_bytes(),
                    bd_address=self.args.bdaddr.lower(),
                    spoof=self.args.spoof,
                    random=(target.address_type == BDAddress.RANDOM)
                )

            # Start our proxy
            self.proxy.start()

            # Set output PCAP file if provided
            if self.args.output is not None:
                pcap_mon = self.proxy.get_pcap_monitor(self.args.output)
                pcap_mon.start()
            else:
                pcap_mon = None

            # Create a Wireshark monitor
            if self.args.wireshark:
                ws_mon = self.proxy.get_wireshark_monitor()
                ws_mon.start()
            else:
                ws_mon = None

            print('Proxy is ready, press a key to stop.')
            input()

            # Stop proxy
            self.proxy.stop()
            self.proxy = None

            # Stop Wireshark monitor
            if ws_mon is not None:
                ws_mon.stop()
                ws_mon.detach()

            # Stop PCAP monitor
            if pcap_mon is not None:
                pcap_mon.stop()
                pcap_mon.detach()


def ble_proxy_main():
    """Launcher for wble-proxy.
    """
    app = BleProxyApp()
    run_app(app)



"""WHAD protocol message abstraction
"""
import struct
import logging

from typing import Any, Callable, Type
from abc import abstractmethod

from scapy.packet import Packet

from whad.protocol.whad_pb2 import Message
from whad.hub.registry import Registry

def dissect_failsafe(f):
    """Failsafe decorator for WHAD's Hub message classes that implement the
    AbstractPacket class.
    """
    def _fail_wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except struct.error:
            logger = logging.getLogger(f.__qualname__)
            logger.debug("Parsing error while converting message %s to packet !", args[0])
            return None
    return _fail_wrapper

class AbstractPacketMeta(type):
    """Hub packet metaclass"""
    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (hasattr(subclass, 'to_packet') and
                callable(subclass.to_packet) and
                hasattr(subclass, 'from_packet') and
                callable(subclass.from_packet))

class AbstractPacket(metaclass=AbstractPacketMeta):
    """Abstract packet class
    """

    @abstractmethod
    def to_packet(self) -> Packet:
        """Convert message to its corresponding packet representation.

        :return: Corresponding Scapy representation
        :rtype: Packets
        """
        return None

class AbstractEventMeta(type):
    """Hub event metaclass
    """
    def __instancecheck__(cls, instance: Any) -> bool:
        return cls.__instancecheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (hasattr(subclass, 'to_event') and
                callable(subclass.to_event) and
                hasattr(subclass, 'from_event') and
                callable(subclass.from_event))

class AbstractEvent(metaclass=AbstractEventMeta):
    """Abstract event class.
    """


class PbField(object):
    """Protocol Buffers field model
    """

    def __init__(self, path: str, field_type, optional=False):
        self.__path = path
        self.__type = field_type
        self.__optional = optional

    def __repr__(self):
        """String representation."""
        return f"PbField(path='{self.__path}', type='{self.__type.__name__}')"

    @property
    def path(self):
        return self.__path

    @property
    def type(self):
        return self.__type

    def is_optional(self):
        """Determine if this field is optional
        """
        return self.__optional

    def update(self, message: Message):
        """Update field in protobuf message
        """
        pass

class PbFieldInt(PbField):
    """Protocol buffers integer field model
    """

    def __init__(self, path: str, optional: bool = False):
        """Create a PB field model for integer.
        """
        super().__init__(path, int, optional=optional)

class PbFieldBytes(PbField):
    """Protocol buffers bytes field model
    """

    def __init__(self, path:  str, optional: bool = False):
        """Create a PB field model for bytes
        """
        super().__init__(path, bytes, optional=optional)

class PbFieldArray(PbField):
    """Protocol buffers array field model
    """

    def __init__(self, path: str, optional: bool = False):
        """Create a PB field model for arrays.
        """
        super().__init__(path, list, optional=optional)

class PbFieldBool(PbField):
    """Protocol buffers bool field model
    """

    def __init__(self, path: str, optional: bool = False):
        """Create a PB field model for bools.
        """
        super().__init__(path, bool, optional=optional)

class PbFieldMsg(PbField):
    """Protocol buffers message field model
    """

    def __init__(self, path: str, wrap_class, optional: bool = False):
        """Create a PB field model for messages.
        """
        super().__init__(path, wrap_class, optional=optional)


class HubMessage(object):
    """Main class from which any ProtocolHub message derives from.
    """

    def __init__(self,  message: Message = None):
        # Allocate a new Protobuf message structure if required
        if message is None:
            self.__msg = Message()
        else:
            self.__msg = message

        # Optional callback method to call when message has been sent
        self.__callback = None

    def callback(self, cb: Callable["HubMessage", int]):
        """Attach a callback to be notified when message is sent, or if an error
        occurred.

        :param handler: Callback function to call when message is sent or if an error
                        occurred.
        :type handler: callable
        """
        self.__callback = cb

    def has_callback(self) -> bool:
        """Determine if message has a callback set.
        """
        return self.__callback is not None

    def sent(self):
        """Notify a successful message transmission.
        """
        if self.__callback is not None:
            self.__callback(self, 0)

    def error(self, error: int):
        """Notify an error while sending message.

        :param error: Error code
        :type error: int
        """
        assert error > 0
        if self.__callback is not None:
            self.__callback(self, error)

    def serialize(self) -> bytes:
        """Serialize hub message.
        """
        return self.__msg.SerializeToString()

    def set_field_value(self, field: PbField, value):
        """Set a message field value.
        """
        path_nodes = field.path.split('.')
        root_node = self.message
        for node in path_nodes[:-1]:
            root_node = getattr(root_node, node)

        if hasattr(root_node, path_nodes[-1]):
            # If we are dealing with a PB array, we cannot set its value but
            # need to call extend() to add our array items.
            if isinstance(value, list):
                getattr(root_node, path_nodes[-1]).extend(value)
            else:
                setattr(root_node, path_nodes[-1], value)
        else:
            raise IndexError()

    def get_field_value(self, field: PbField):
        """Get a message field value.
        """
        # Walk to the penultimate field
        path_nodes = field.path.split('.')
        root_node = self.message

        for node in path_nodes[:-1]:
            if hasattr(root_node, node):
                root_node = getattr(root_node, node)
            else:
                raise IndexError()

        # Return the final field
        if hasattr(root_node, path_nodes[-1]):
            if isinstance(field, PbFieldMsg):
                return field.type(getattr(root_node, path_nodes[-1]))
            else:
                if field.is_optional() and not root_node.HasField(path_nodes[-1]):
                    return None
                else:
                    return getattr(root_node, path_nodes[-1])
        else:
            raise IndexError()

    @property
    def message(self):
        """Underlying protobuf message
        """
        return self.__msg


class pb_bind(object):
    """Decorator to add a versioned subclass to a registry.
    """

    def __init__(self, registry: Type[Registry], name: str, version: int):
        self.__registry = registry
        self.__name = name
        self.__version = version

    def __call__(self, clazz):
        """Decorate our class
        """
        # Register our class with the corresponding name
        self.__registry.add_node_version(self.__version, self.__name, clazz)

        # Add to our class a category type corresponding to the registry name
        clazz.message_type = self.__registry.NAME
        clazz.message_name = self.__name
        return clazz


class PbMessageWrapper(HubMessage):
    """Protocol Buffers message wrapper

    This class allows a mapping between its class fields (declared as `PbField`
    objects) and its underlying Protocol Buffers message object, in order to
    transparently access and updates these fields through simple parameters.
    """

    def __init__(self, message: Message = None, **kwargs):
        """Initialize a `PbMessageWrapper` object.

        This code goes through all the declared properties and finds out the
        ones based on `PbField` and then automatically binds them to the
        corresponding parameter name.
        """
        self.__pb_fields = {}

        # Create our HubMessage
        super().__init__(message=message)

        # Browse our properties and list PB fields
        for prop in dir(self):
            if not prop.startswith('_'):
                try:
                    prop_obj = getattr(self, prop)
                    if not callable(prop_obj):
                        # check field type
                        if isinstance(prop_obj, PbField):
                            self.__pb_fields[prop] = prop_obj
                except AttributeError:
                    pass

        # Override message values with keyword arguments
        for message_field in kwargs:
            if message_field in self.__pb_fields:
                self.set_field_value(self.__pb_fields[message_field], kwargs[message_field])

    def __getattribute__(self, name):
        """Override the class __getattribute__ function to allow transparent
        access to protocol buffers fields for the given message and corresponding
        field paths.
        """
        if name.startswith('_'):
            return object.__getattribute__(self, name)
        elif name in self.__pb_fields:
            return self.get_field_value(self.__pb_fields[name])
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        """Set underlying message fields given their paths with a given value.
        """
        if name.startswith('_'):
            self.__dict__[name] = value
        elif name in self.__pb_fields:
            self.set_field_value(self.__pb_fields[name], value)

    def __repr__(self):
        """Generate a string representation of our message
        """
        fields = []
        for prop_name in self.__pb_fields:
            value = self.get_field_value(self.__pb_fields[prop_name])
            if value is not None:
                fields.append((prop_name, value))

        return f"<{self.__class__.__module__}.{self.__class__.__name__} " + ', '.join(
            [f'{name}={value}' for name, value in fields]
        ) + ">"

    @classmethod
    def parse(parent_class, version: int, message: Message):
        """Parse a generic protobuf message message.
        """
        return parent_class(message=message)



"""WHAD Protocol Dot15d4 address messages abstraction layer.
"""

from ..message import pb_bind, PbFieldInt, PbFieldBytes, PbMessageWrapper
from . import Dot15d4Domain

@pb_bind(Dot15d4Domain, 'set_node_addr', 1)
class SetNodeAddress(PbMessageWrapper):
    """Device info query message class
    """

    address = PbFieldBytes('dot15d4.set_node_addr.address')
    addr_type = PbFieldInt('dot15d4.set_node_addr.address_type')



'''Whad generic protocol stack
'''

from .layer import source, alias, state, instance, LayerState,  Layer, ContextualLayer


__all__ = [
    'Layer',
    'source',
    'instance',
    'alias',
    'state',
    'LayerState',
    'ContextualLayer'
]



from whad.helpers import bd_addr_to_bytes, asciiz, is_message_type, swap_bits
from whad.protocol.whad_pb2 import Message
from tests.sample_messages import DISCOVERY_SAMPLE_MESSAGES,GENERIC_SAMPLE_MESSAGES, BLE_SAMPLE_MESSAGES
from google.protobuf.text_format import Parse
import pytest

@pytest.mark.parametrize("test_input, expected", [
    ("11:22:33:44:55:66", b"\x66\x55\x44\x33\x22\x11"),
    ("aA:Bb:cC:dd:EE:fF", b"\xff\xee\xdd\xcc\xbb\xaa"),
    ("ff:ff:ff:FF:FF:FF", b"\xff\xff\xff\xff\xff\xff"),
    ("ff  :  ff : ff  :F F    :F  F:  F F", None),
    ("test", None),
    ("", None),
    (42, None)

])
def test_bd_addr_to_bytes(test_input, expected):
    assert bd_addr_to_bytes(test_input) == expected

@pytest.mark.parametrize("test_input, expected", [
    (b"\x41\x42\x43\x44", "ABCD"),
    (b"\x31\x62\x33\x64\x35\x66", "1b3d5f"),
    (b"", ""),
    (None, None),
    (42, None)
])
def test_asciiz(test_input, expected):
    assert asciiz(test_input) ==  expected

@pytest.mark.parametrize("test_input, expected",
    [((msg_type, "discovery", msg),True) for msg_type, msg in DISCOVERY_SAMPLE_MESSAGES.items()] +
    [((msg_type, "ble", msg),False) for msg_type, msg in DISCOVERY_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "discovery", msg),False) for msg_type, msg in DISCOVERY_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "ble", msg),False) for msg_type, msg in DISCOVERY_SAMPLE_MESSAGES.items()] +

    [((msg_type, "generic", msg),True) for msg_type, msg in GENERIC_SAMPLE_MESSAGES.items()] +
    [((msg_type, "discovery", msg),False) for msg_type, msg in GENERIC_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "generic", msg),False) for msg_type, msg in GENERIC_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "discovery", msg),False) for msg_type, msg in GENERIC_SAMPLE_MESSAGES.items()] +

    [((msg_type, "ble", msg),True) for msg_type, msg in BLE_SAMPLE_MESSAGES.items()] +
    [((msg_type, "generic", msg),False) for msg_type, msg in BLE_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "ble", msg),False) for msg_type, msg in BLE_SAMPLE_MESSAGES.items()] +
    [(("wrong_type", "generic", msg),False) for msg_type, msg in BLE_SAMPLE_MESSAGES.items()]

    )

def test_is_message_type(test_input, expected):
    msg_type, category, message = test_input
    msg = Message()
    Parse(message, msg)
    assert is_message_type(msg, category, msg_type) ==  expected

@pytest.mark.parametrize("test_input, expected", [
    (0x48, 0x12),
    (b"\xF0\xF0\xF0\xF0", b"\x0F\x0F\x0F\x0F"),
    (b"\x12\x34", b"\x48\x2c"),
    ("toto", None)
    ])
def test_swap_bits(test_input, expected):
    assert swap_bits(test_input) == expected

#TODO: message_filter test
#TODO: is_message_msg_type test



from whad.helpers import list_domains
from whad.protocol.phy.phy_pb2 import SF7, SF8, SF9, SF10, SF11, SF12, \
    CR45, CR46, CR47, CR48
from whad.hub.phy import LoRaCodingRate, LoRaSpreadingFactor
from whad.phy.exceptions import InvalidParameter

LORA_SF_MAP = {
    7: LoRaSpreadingFactor.SF7,
    8: LoRaSpreadingFactor.SF8,
    9: LoRaSpreadingFactor.SF9,
    10: LoRaSpreadingFactor.SF10,
    11: LoRaSpreadingFactor.SF11,
    12: LoRaSpreadingFactor.SF12
}

LORA_CR_MAP = {
    45: LoRaCodingRate.CR45,
    46: LoRaCodingRate.CR46,
    47: LoRaCodingRate.CR47,
    48: LoRaCodingRate.CR48
}

def get_physical_layers_by_domain(domain):
    '''
    Returns physical layer of a specific domain. Returns None if no physical layer found.
    '''
    try:
        phys = __import__("whad.{}".format(domain),  globals(), locals(), ["PHYS"])
        return vars(phys)["PHYS"]
    except (ImportError, KeyError):
        return None
    
def get_all_physical_layers():
    '''
    Returns all available physical layers.
    '''
    available_phys = {}
    for domain in list_domains():
        phys = get_physical_layers_by_domain(domain)
        if phys is not None:
            available_phys.update(phys)
    return available_phys

def lora_sf(sf: int=7):
    '''
    Convert LoRa spreading factor value into WHAD compatible constant
    '''
    if sf in LORA_SF_MAP:
        return LORA_SF_MAP[sf]
    else:
        raise InvalidParameter('spreading factor')
    
def lora_cr(cr: int=45):
    '''
    Convert LoRa coding rate value into WHAD compatible constant
    '''
    if cr in LORA_CR_MAP:
        return LORA_CR_MAP[cr]
    else:
        raise InvalidParameter('coding rate')




"""WHAD Protocol 802.15.4 domain message abstraction layer.
"""
from typing import List
from dataclasses import dataclass, field, fields

from scapy.config import conf
from scapy.layers.dot15d4 import Dot15d4FCS

from whad.scapy.layers.dot15d4tap import Dot15d4TAP_Hdr, Dot15d4TAP_TLV_Hdr,\
    Dot15d4TAP_Received_Signal_Strength, Dot15d4TAP_Channel_Assignment, \
    Dot15d4TAP_Channel_Center_Frequency, Dot15d4TAP_Link_Quality_Indicator, \
    Dot15d4TAP_FCS_Type

from whad.protocol.dot15d4.dot15d4_pb2 import Dot15d4MitmRole, AddressType
from whad.hub.registry import Registry
from whad.hub.message import HubMessage, pb_bind
from whad.hub.metadata import Metadata, channel_to_frequency
from whad.hub import ProtocolHub

class Commands:
    """Dot15d4 commands
    """
    SetNodeAddress = 0x00
    Sniff = 0x01
    Jam = 0x02
    EnergyDetection = 0x03
    Send = 0x04
    SendRaw = 0x05
    EndDeviceMode = 0x06
    CoordinatorMode = 0x07
    RouterMode = 0x08
    Start = 0x09
    Stop = 0x0a
    ManInTheMiddle = 0x0b

class MitmRole:
    """Dot15d4 Mitm role
    """
    REACTIVE=Dot15d4MitmRole.REACTIVE_JAMMER
    CORRECTOR=Dot15d4MitmRole.CORRECTOR

class NodeAddressType:
    """IEEE 802.15.4 Node address type enum.
    """
    SHORT=AddressType.SHORT
    EXTENDED=AddressType.EXTENDED

class NodeAddress(object):
    """IEEE 802.15.4 Node address.
    """

    def __init__(self, address: int, addr_type: int):
        self.__address = address
        self.__address_type = addr_type

    @property
    def address(self) -> int:
        """Node address"""
        return self.__address

    @property
    def address_type(self) -> int:
        """Address type"""
        return self.__address_type

class NodeAddressShort(NodeAddress):
    """Dot15d4 short node address

    This class reprensents a short 802.15.4 node address (16 bits)
    """

    def __init__(self, address: int):
        """Initialize a node short address.
        """
        assert address >= 0
        assert address < 0x10000
        super().__init__(address, NodeAddressType.SHORT)

class NodeAddressExt(NodeAddress):
    """Dot15d4 extended node address

    This class represents an extended 802.15.4 node address (64 bits).
    """

    def __init__(self, address: int):
        """Initialize a node extended address.

        :param address: a 64-bit node address
        """
        assert address >= 0
        assert address <= 0x10000000000000000
        super().__init__(address, NodeAddressType.EXTENDED)


@dataclass(repr=False)
class Dot15d4Metadata(Metadata):
    """Dot15d4 meta-data holding class.
    """
    is_fcs_valid : bool = None
    lqi : int = None
    timestamp : int = None

    def convert_to_header(self) -> Dot15d4TAP_Hdr:
        """Convert stored metadata into a Scapy Dot15d4TAP_Hdr instance.

        :return: Scapy header for 802.15.4 packet
        :rtype: Dot15d4TAP_Hdr
        """
        timestamp = None
        tlv = []
        if self.timestamp is not None:
            timestamp = self.timestamp
        if self.rssi is not None:
            tlv.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_Received_Signal_Strength(rss = self.rssi))
        if self.lqi is not None:
            tlv.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_Link_Quality_Indicator(lqi = self.lqi))
        if self.channel is not None:
            tlv.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_Channel_Assignment(channel_number=self.channel,
                                                                          channel_page=0))
            channel_frequency = channel_to_frequency(self.channel) * 1000
            tlv.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_Channel_Center_Frequency(channel_frequency=channel_frequency))
        return Dot15d4TAP_Hdr(data=tlv), timestamp

    @classmethod
    def convert_from_header(cls, pkt) -> "Dot15d4Metadata":
        """Load metadata from packet header into a Dot15d4Metadata structure
        """
        # Default values for DLTs that do not have any metadata.
        rssi = 0
        lqi = 200
        channel = 15

        # Packets from PCAP with DLT 283 (IEEE 802.15.4 TAP)
        # Other DLTs don't have any metadata about RSSI, LQI or channel.
        if Dot15d4TAP_Hdr in pkt:
            for layer in pkt[Dot15d4TAP_Hdr].data:
                if Dot15d4TAP_Received_Signal_Strength in layer:
                    rssi = layer.rss
                elif Dot15d4TAP_Link_Quality_Indicator in layer:
                    lqi = layer.lqi
                elif Dot15d4TAP_Channel_Assignment in layer:
                    channel = layer.channel_number
                else:
                    pass

        return Dot15d4Metadata(
            rssi = int(rssi),
            lqi = lqi,
            channel = channel,
            timestamp = int(100000 * pkt.time)
        )

def generate_dot15d4_metadata(message: HubMessage) -> Dot15d4Metadata:
    """Generate a Dot15d4Metadata object from a WHAD message.
    """
    metadata = Dot15d4Metadata()

    if message.lqi is not None:
        metadata.lqi = message.lqi
    if message.rssi is not None:
        metadata.rssi = message.rssi
    metadata.channel = message.channel
    if message.timestamp is not None:
        metadata.timestamp = message.timestamp
    if message.fcs_validity is not None:
        metadata.is_fcs_valid = message.fcs_validity

    return metadata

@pb_bind(ProtocolHub, name="dot15d4", version=1)
class Dot15d4Domain(Registry):
    """WHAD Dot15d4 domain messages parser/factory.
    """

    NAME = 'dot15d4'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a Dot15d4 domain instance
        """
        self.proto_version = version

    def is_packet_compat(self, packet) -> bool:
        """Determine if a packet is a Dot15d4 packet.
        """
        return isinstance(packet.metadata, Dot15d4Metadata)

    def convert_packet(self, packet) -> HubMessage:
        """Convert a Dot15d4 packet to SendPdu or SendBlePdu message.
        """
        if isinstance(packet.metadata, Dot15d4Metadata):
            if packet.metadata.raw:
                return Dot15d4Domain.bound('send_raw', self.proto_version).from_packet(
                    packet, channel=packet.metadata.channel
                )
            else:
                return Dot15d4Domain.bound('send', self.proto_version).from_packet(
                    packet, channel=packet.metadata.channel
                )
        else:
            # Error
            return None

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if hasattr(packet, "metadata"):
            header, timestamp = packet.metadata.convert_to_header()
        else:
            header = Dot15d4TAP_Hdr()
            timestamp = None

        header.data.append(Dot15d4TAP_TLV_Hdr()/Dot15d4TAP_FCS_Type(
            fcs_type=int(Dot15d4FCS in packet)
            )
        )
        formatted_packet = header/packet
        return formatted_packet, timestamp

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD Dot15d4 Domain message as seen by protobuf
        """
        message_type = message.dot15d4.WhichOneof('msg')
        message_clazz = Dot15d4Domain.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def create_set_node_address(self, address: NodeAddress) -> HubMessage:
        """Create a SetNodeAddress message.

        :param address: instance of `Dot15d4Address`
        :type address: NodeAddress
        :return: instance of `SetNodeAddress`
        """
        return Dot15d4Domain.bound('set_node_addr', self.proto_version)(
            address=address.address,
            addr_type=address.address_type
        )

    def create_sniff_mode(self, channel: int) -> HubMessage:
        """Create a SniffMode message

        :param channel: Channel to sniff
        :type channel: int
        :return: instance of `SniffMode`
        """
        return Dot15d4Domain.bound('sniff', self.proto_version)(
            channel=channel
        )

    def create_jam_mode(self, channel: int) -> HubMessage:
        """Create a JamMode message

        :param channel: Channel to jam
        :type channel: int
        :return: instance of `JamMode`
        """
        return Dot15d4Domain.bound('jam', self.proto_version)(
            channel=channel
        )

    def create_energy_detection_mode(self, channel: int) -> HubMessage:
        """Create a EnergyDetectionMode message

        :param channel: Channel to detect
        :type channel: int
        :return: instance of `EnergyDetectionMode`
        """
        return Dot15d4Domain.bound('ed', self.proto_version)(
            channel=channel
        )

    def create_end_device_mode(self, channel: int) -> HubMessage:
        """Create a EndDeviceMode message

        :param channel: Channel to use for end device
        :type channel: int
        :return: instance of `EndDeviceMode`
        """
        return Dot15d4Domain.bound('end_device', self.proto_version)(
            channel=channel
        )

    def create_router_mode(self, channel: int) -> HubMessage:
        """Create a RouterMode message

        :param channel: Channel to use for router
        :type channel: int
        :return: instance of `RouterMode`
        """
        return Dot15d4Domain.bound('router', self.proto_version)(
            channel=channel
        )

    def create_coord_mode(self, channel: int) -> HubMessage:
        """Create a CoordMode message

        :param channel: Channel to use for router
        :type channel: int
        :return: instance of `CoordMode`
        """
        return Dot15d4Domain.bound('coordinator', self.proto_version)(
            channel=channel
        )

    def create_mitm_mode(self, role: int) -> HubMessage:
        """Create a MitmMode message

        :return: instance of `MitmMode`
        """
        return Dot15d4Domain.bound('mitm', self.proto_version)(
            role=role
        )

    def create_start(self) -> HubMessage:
        """Create a Start message

        :return: instance of `Start`
        """
        return Dot15d4Domain.bound('start', self.proto_version)()

    def create_stop(self) -> HubMessage:
        """Create a Stop message

        :return: instance of `Stop`
        """
        return Dot15d4Domain.bound('stop', self.proto_version)()

    def create_send_pdu(self, channel: int, pdu: bytes) -> HubMessage:
        """Create a SendPdu message

        :param channel: Channel on which the PDU has to be sent
        :type channel: int
        :param pdu: PDU to send
        :type pdu: bytes
        :return: instance of `SendPdu`
        """
        return Dot15d4Domain.bound('send', self.proto_version)(
            channel=channel,
            pdu=pdu
        )

    def create_send_raw_pdu(self, channel: int, pdu: bytes, fcs: int) -> HubMessage:
        """Create a SendPdu message

        :param channel: Channel on which the PDU has to be sent
        :type channel: int
        :param pdu: PDU to send
        :type pdu: bytes
        :param fcs: Frame check sequence
        :type fcs: int
        :return: instance of `SendPdu`
        """
        return Dot15d4Domain.bound('send_raw', self.proto_version)(
            channel=channel,
            pdu=pdu,
            fcs=fcs
        )

    def create_jammed(self, timestamp: int) -> HubMessage:
        """Create a jammed notification.

        :param timestamp: Timestamp when jamming is successful
        :type timestamp: int
        :return: instance of `Jammed`
        """
        return Dot15d4Domain.bound('jammed', self.proto_version)(
            timestamp=timestamp
        )

    def create_energy_detection_sample(self, timestamp: int, sample: int) -> HubMessage:
        """Create an energy detection sample notification message.

        :param timestamp: Timestamp at wich the sample has been computed
        :type timestamp: int
        :param sample: Computed sample
        :type sample: int
        :return: instance of `EnergyDetectionSample`
        """
        return Dot15d4Domain.bound('ed_sample', self.proto_version)(
            timestamp=timestamp,
            sample=sample
        )

    def create_raw_pdu_received(self, channel: int, pdu: bytes, fcs: int, rssi: int = None, \
                             timestamp: int = None, fcs_validity: bool = None, \
                             lqi: int = None):
        """Create a received PDU notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Received PDU
        :type pdu: bytes
        :param fcs: Frame Check Sequence
        :type fcs: int
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Timestamp at which the PDU has been received
        :type timestamp: int, optional
        :param fcs_validity: Specify if the FCS field is valid or not
        :type fcs_validity: bool, optional
        :param lqi: Link Quality indicator
        :type lqi: int, optional
        :return: instance of `RawPduReceived`
        """
        # Create our RawPduReceived message with mandatory fields
        msg = Dot15d4Domain.bound('raw_pdu', self.proto_version)(
            channel=channel,
            pdu=pdu,
            fcs=fcs
        )

        # Add optional fields if they are provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if fcs_validity is not None:
            msg.fcs_validity = fcs_validity
        if lqi is not None:
            msg.lqi = lqi

        # Return the generated message
        return msg

    def create_pdu_received(self, channel: int, pdu: bytes, rssi: int = None, \
                             timestamp: int = None, fcs_validity: bool = None, \
                             lqi: int = None):
        """Create a received PDU notification message.

        :param channel: Channel on which the PDU has been received
        :type channel: int
        :param pdu: Received PDU
        :type pdu: bytes
        :param rssi: Received signal strength indicator
        :type rssi: int, optional
        :param timestamp: Timestamp at which the PDU has been received
        :type timestamp: int, optional
        :param fcs_validity: Specify if the FCS field is valid or not
        :type fcs_validity: bool, optional
        :param lqi: Link Quality indicator
        :type lqi: int, optional
        :return: instance of `RawPduReceived`
        """
        # Create our PduReceived message with mandatory fields
        msg = Dot15d4Domain.bound('pdu', self.proto_version)(
            channel=channel,
            pdu=pdu,
        )

        # Add optional fields if they are provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp
        if fcs_validity is not None:
            msg.fcs_validity = fcs_validity
        if lqi is not None:
            msg.lqi = lqi

        # Return the generated message
        return msg

@pb_bind(ProtocolHub, name="rf4ce", version=1)
class RF4CEDomain(Dot15d4Domain):
    NAME = 'rf4ce'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a RF4CE domain instance
        """
        super().__init__(version)
        conf.dot15d4_protocol = "rf4ce"


@pb_bind(ProtocolHub, name="zigbee", version=1)
class ZigBeeDomain(Dot15d4Domain):
    NAME = 'zigbee'
    VERSIONS = {}

    def __init__(self, version: int):
        """Initializes a ZigBee domain instance
        """
        super().__init__(version)
        conf.dot15d4_protocol = "zigbee"


from .address import SetNodeAddress
from .mode import SniffMode, RouterMode, EndDeviceMode, CoordMode, EnergyDetectionMode, \
    JamMode, MitmMode, Start, Stop, Jammed, EnergyDetectionSample
from .pdu import SendPdu, SendRawPdu, PduReceived, RawPduReceived

__all__ = [
    'SetNodeAddress',
    'SniffMode',
    'RouterMode',
    'EndDeviceMode',
    'CoordMode',
    'EnergyDetectionMode',
    'JamMode',
    'MitmMode',
    'Start',
    'Stop',
    'SendPdu',
    'SendRawPdu',
    'PduReceived',
    'RawPduReceived',
    'Jammed',
    'EnergyDetectionSample',
    'Dot15d4Domain',
    'NodeAddress',
    'NodeAddressShort',
    'NodeAddressExt',
    'NodeAddressType',
    'MitmRole'
]



from .base import Zigbee
from .sniffer import Sniffer
from .enddevice import EndDevice
from .coordinator import Coordinator

__all__ = [
    "Zigbee",
    "Sniffer",
    "EndDevice",
    "Coordinator"
]



"""
Bluetooth Low Energy GATT procedure

This class must be inherited to implement a GATT procedure.

# Procedure states

A GATT procedure has 3 predefined internal states:
- STATE_INITIAL: default state for a procedure
- STATE_DONE: state is reached when the GATT procedure has finished without error
- STATE_ERROR: state is reached when an unrecoverable error occurred

User-defined states could be defined starting with the base user state value
STATE_USER.

The `done()` method can be used to check if the procedure has finished and the `error()`
method to assess if the procedure is in error.

# Procedure trigger mechanism

The `trigger()` method must be overriden by inherited classes to provide the caller a
way to check if the procedure needs to be triggered. An ATT read request procedure
will check for ATT ReadRequest packets and would return `True` if such a request is
received. When the `trigger()` method returns `True`, the tiny BLE stack will consider
this procedure active and will call its `process_request()` method each time a packet is
received, until the procedure is over (STATE_DONE).

# Packet processing

Any procedure class that inherits from `Procedure` must implement its own `process_request()`
method that must return a list of packets to be sent back to the requesting peer. Since this
tiny stack has been designed for unit-testing, there is no mechanism to trigger any delayed
packet transmission. Delayed packets must be returned as well in the list of packets, and will
be processed as soon as possible.

"""
import logging
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response

from whad.device.mock.procedure import StackProcedure, ProcedureState, UnexpectedProcError

from .attribute import Attribute

logger = logging.getLogger(__name__)

class BleClientProcedure(StackProcedure):
    """Bluetooth Low Energy ATT Client procedure."""

    # ATT Operation code (0 is invalid)
    OPCODE = 0

    class States(ProcedureState):
        """BLE ATT Client states."""
        SUB_DONE = ProcedureState.USER

    def __init__(self, packets: List[Packet]):
        super().__init__(packets)

    @property
    def states(self):
        return BleClientProcedure.States

    def done(self) -> bool:
        """
        Determine if the procedure has completed.

        TODO: replace with call to success() in other dependencies."""
        return self.success()

    def att_error_response(self, handle: int, ecode: int) -> list[Packet]:
        """
        Generate an ATT error response.

        :param handle: Attribute handle
        :type handle: int
        :param ecode: ATT error code as defined in Vol 3, Part F, section 3.3.3, Table 3.3
        :type ecode: int
        """
        return [ATT_Error_Response(request=self.OPCODE, handle=handle, ecode=ecode)]

    def process_request(self, request: Packet) -> List[Packet]:
        """Process a received ATT request/response.

        :return: List of packets (PDUs) to send once the received PDUs processed.
        :rtype: List
        :raise UnexpectedProcError: An unexpected error occurred while processing incoming packets.
        """
        raise UnexpectedProcError(request)

class BleServerProcedureState(ProcedureState):
    """States specific to BLE server write procedure. """
    INDICATION_SENT = ProcedureState.USER

class BleServerProcedure(StackProcedure):
    """Bluetooth Low Energy ATT Server procedure."""

    # ATT Operation code (0 is invalid)
    OPCODE = 0

    @classmethod
    def trigger(cls, request) -> bool:
        """Trigger or not the procedure."""
        logger.warning("[ble::stack::mock::Procedure] trigger method must be overriden")
        logger.debug("[ble::stack::mock::Procedure] trigger() called with packet %s", bytes(request).hex())
        return False


    def __init__(self, attributes: list, mtu: int = 23):
        """Initialization.

        :param  attributes: GATT server attributes
        :param  attributes: dict
        :param  mtu: Maximum transmission unit
        :type   mtu: int, optional
        """
        # Call parent init method.
        super().__init__([])

        # Save attribute list
        self.__attributes = attributes

        # Save ATT MTU
        self.__mtu = mtu

    @property
    def attributes(self) -> list[Attribute]:
        """GATT attributes dictionary."""
        return self.__attributes

    @property
    def mtu(self) -> int:
        """ATT MTU."""
        return self.__mtu

    @property
    def states(self):
        """Specify the states enum for this procedure."""
        return BleServerProcedureState

    def done(self) -> bool:
        """
        Determine if the procedure has completed.

        TODO: replace with call to success() in other dependencies."""
        return self.success()

    def att_error_response(self, handle: int, ecode: int) -> list[Packet]:
        """
        Generate an ATT error response.

        :param handle: Attribute handle
        :type handle: int
        :param ecode: ATT error code as defined in Vol 3, Part F, section 3.3.3, Table 3.3
        :type ecode: int
        """
        return [ATT_Error_Response(request=self.OPCODE, handle=handle, ecode=ecode)]

    def process_request(self, request: Packet) -> List[Packet]:
        """Process a received ATT request/response.

        :return: List of packets (PDUs) to send once the received PDUs processed.
        :rtype: List
        :raise UnexpectedProcError: An unexpected error occurred while processing incoming packets.
        """
        raise UnexpectedProcError(request)




from whad.phy.connector import Phy, Endianness, TxPower
from whad.phy.utils.definitions import ASKModulationScheme, FSKModulationScheme, GFSKModulationScheme, \
    PSKModulationScheme, BPSKModulationScheme, QPSKModulationScheme, OQPSKModulationScheme, \
    OOKModulationScheme, PhysicalLayer

__all__ = [
    'Phy',
    'Endianness',
    'TxPower', 
    'OOKModulationScheme',
    'ASKModulationScheme',
    'FSKModulationScheme',
    'GFSKModulationScheme',
    'PSKModulationScheme',
    'BPSKModulationScheme',
    'QPSKModulationScheme',
    'OQPSKModulationScheme',
    'PhysicalLayer'
]



"""
WHAD Enhanced ShockBurst base connector

This module provides the `ESB` connector class to interact with Enhanced
ShockBurst compatible devices.
"""
from scapy.packet import Packet

from whad.device import Connector
from whad.helpers import message_filter
from whad.exceptions import UnsupportedDomain, UnsupportedCapability

from whad.scapy.layers.esb import ESB_Hdr,ESB_Payload_Hdr,ESB_Ack_Response

from whad.hub.discovery import Domain, Capability
from whad.hub.esb import ESBMetadata, EsbNodeAddress, Commands
from whad.hub.generic.cmdresult import Success, CommandResult
from whad.hub.events import JammedEvt

from whad.esb.esbaddr import ESBAddress

class ESB(Connector):
    """
    Enhanced ShockBurst protocol connector.

    This connector drives a Enhanced ShockBurst (ESB) capable device with
    ESB-specific WHAD messages. It is required by various role classes to
    interact with a real device and pre-process domain-specific messages.
    """

    domain = "esb"

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        if isinstance(packet, bytes):
            packet = ESB_Hdr(packet)
        return self.hub.esb.format(packet)

    def __init__(self, device=None, synchronous=False):
        """
        Initialize the connector, open the device (if not already opened), discover
        the services (if not already discovered).

        If `synchronous` is set to True, PDUs must be processed manually and
        won't be forwarded to PDU-related callbacks.
        """
        self.__ready = False
        super().__init__(device)

        # Metadata cache
        self.__cached_channel = None
        self.__cached_address = None

        # Capability cache
        self.__can_send = None
        self.__can_send_raw = None

        # Open device and make sure it is compatible
        self.device.open()
        self.device.discover()

        # Check if device supports Enhanced ShockBurst
        if not self.device.has_domain(Domain.Esb):
            raise UnsupportedDomain("ESB")

        # Mark connector as ready
        self.__ready = True

        # Set synchronous mode
        self.enable_synchronous(synchronous)

    def close(self):
        """Stop the WHAD adpater and close its related device.
        """
        self.stop()
        self.device.close()

    def can_sniff(self):
        """
        Determine if the device implements a sniffer mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.Sniff)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def can_send(self):
        """
        Determine if the device can transmit packets.
        """
        if self.__can_send is None:
            commands = self.device.get_domain_commands(Domain.Esb)
            self.__can_send = ((commands & (1 << Commands.Send))>0 or
                               (commands & (1 << Commands.SendRaw)))
        return self.__can_send

    def send(self,pdu, address=None, channel=None, retransmission_count=1):
        """
        Send Enhanced ShockBurst packets (on a single channel).
        """
        if self.can_send():
            # If we don't have address or channels, use the cached ones
            tx_address = address if address is not None else self.__cached_address
            tx_channel = channel if channel is not None else self.__cached_channel
            if self.support_raw_pdu():
                # If we support raw PDU but only got a payload, build a packet
                if ESB_Hdr not in pdu:
                    packet = ESB_Hdr(address=tx_address) / pdu
                else:
                    packet = pdu
            # if we don't support raw PDU and got a packet, crop to keep only the payload
            elif ESB_Hdr in pdu:
                packet = pdu[ESB_Payload_Hdr:]
            # if we don't support raw PDU and got a payload, keep it as it is
            else:
                packet = pdu

            # Generate TX metadata
            packet.metadata = ESBMetadata()
            packet.metadata.channel = tx_channel
            packet.metadata.address = tx_address
            packet.metadata.retransmission_count = retransmission_count
            if self.support_raw_pdu():
                packet.metadata.raw = True
                if bytes.fromhex(packet.address[:2])[0] >= 0x80:
                    packet.preamble = 0xAA
                else:
                    packet.preamble = 0x55

            # Send packet
            return super().send_packet(packet)

        # Cannot send, return False.
        return False

    def support_raw_pdu(self):
        """
        Determine if the device supports raw PDU.
        """
        if self.__can_send_raw is None:
            capabilities = self.device.get_domain_capability(Domain.Esb)
            self.__can_send_raw = not capabilities & Capability.NoRawData
        return self.__can_send_raw


    def can_set_node_address(self):
        """
        Determine if the device can configure a Node address.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.SetNodeAddress)) > 0
        )

    def start_sniff(self, channel : int = None, address : str = "FF:FF:FF:FF:FF",
              show_acknowledgements : bool = False):
        """
        Sniff Enhanced ShockBurst packets.

        :param channel: Channel to listen, None to iterate over all possible channels
        :type channel: int
        :param address: Device address to target
        :type address: str
        :param show_acknowledgements: Sniff packets acknowledgements if set to True (default: False)
        :type show_acknowledgements: bool
        """
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

        self.__cached_address = ESBAddress(address)
        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            self.__cached_channel = channel

        msg = self.hub.esb.create_sniff_mode(
            EsbNodeAddress(self.__cached_address.value),
            channel,
            show_acknowledgements
        )
        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)



    def can_be_prx(self):
        """
        Determine if the device implements a Primary Receiver (PRX) role mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.PrimaryReceiverMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_prx_mode(self, channel):
        """
        Enable Enhanced ShockBurst primary receiver (PRX) mode.
        """
        if not self.can_be_prx():
            raise UnsupportedCapability("PrimaryReceiverMode")

        # Check that we got a channel provided
        if channel is None:
            return False

        # Keep provided channel in cache
        self.__cached_channel = channel

        # Create a PrxMode message.
        msg = self.hub.esb.create_prx_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)


    def can_be_ptx(self):
        """
        Determine if the device implements a Primary Transmitter (PTX) role mode.
        """
        commands = self.device.get_domain_commands(Domain.Esb)
        return (
            (commands & (1 << Commands.PrimaryTransmitterMode)) > 0 and
            (commands & (1 << Commands.Start))>0 and
            (commands & (1 << Commands.Stop))>0
        )

    def enable_ptx_mode(self, channel):
        """
        Enable Enhanced ShockBurst primary transmitter (PTX) mode.
        """
        if not self.can_be_ptx():
            raise UnsupportedCapability("PrimaryTransmitterMode")


        if channel is None:
            # Enable scanning mode
            channel = 0xFF
        else:
            # Keep channel in cache
            self.__cached_channel = channel

        # Create a PtxMode message.
        msg = self.hub.esb.create_ptx_mode(channel)

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def set_node_address(self, address):
        """
        Enable Enhanced ShockBurst primary receiver (PRX) mode.
        """
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        node_address = ESBAddress(address)
        # Keep address in cache
        self.__cached_address = node_address

        # Create a SetNodeAddress message
        msg = self.hub.esb.create_set_node_address(
            EsbNodeAddress(node_address.value)
        )

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def start(self):
        """
        Start currently enabled mode.
        """
        # Create a Start message.
        msg = self.hub.esb.create_start()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def stop(self):
        """
        Stop currently enabled mode.
        """
        # Create a Stop message.
        msg = self.hub.esb.create_stop()

        resp = self.send_command(msg, message_filter(CommandResult))
        return isinstance(resp, Success)

    def on_discovery_msg(self, message):
        pass

    def on_generic_msg(self, message):
        pass

    def on_domain_msg(self, domain, message):
        pass

    def on_packet(self, packet):
        """Incoming packet callback.
        """
        if not self.__ready:
            return

        # Dispatch packet
        if packet.metadata.raw:
            self.on_raw_pdu(packet)
        else:
            self.on_pdu(packet)

    def on_event(self, event):
        """Process incoming events.
        """
        if not self.__ready:
            return

        if isinstance(event, JammedEvt):
            self.on_jammed(event.timestamp)

    def on_raw_pdu(self, packet: Packet):
        """Callback to handle incoming raw PDU

        :param packet: Incoming PDU
        :type packet: Packet
        """
        # Extract the PDU from raw packet
        if ESB_Payload_Hdr in packet:
            pdu = packet[ESB_Payload_Hdr:]
        else:
            pdu = ESB_Payload_Hdr()/ESB_Ack_Response()

        # Propagate metadata to PDU
        pdu.metadata = packet.metadata
        self.on_pdu(pdu)

    def on_pdu(self, packet):
        """Callback for incoming PDU.

        For now, all compatible devices report raw PDUs when it comes to ESB,
        so this should not be called.
        """

    def on_jammed(self, timestamp: int):
        """Jammed event handler.
        """
